# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import csv
from datetime import datetime
import logging
import math
import os
import re
import struct
from typing import (
    Any, Collection, Dict, Iterable, List, Optional, Sequence, Set, Tuple,
    TYPE_CHECKING, Union)

import numpy
from numpy import float64, floating, integer, uint8, uint32
from numpy.typing import NDArray
import quantities
import neo  # type: ignore[import]

from spinn_utilities.log import FormatAdapter
from spinnman.messages.eieio.data_messages import EIEIODataHeader


from pacman.model.graphs.common import Slice, MDSlice
from pacman.utilities.utility_calls import get_keys

from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utility_models import (
    ReverseIPTagMulticastSourceMachineVertex)
from spinn_front_end_common.interface.buffer_management.storage_objects \
    import BufferDatabase
from spinn_front_end_common.utilities.base_database import _SqliteTypes
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_WORD, BITS_PER_WORD)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.types import ViewIndices
from spynnaker.pyNN.utilities.buffer_data_type import BufferDataType
from spynnaker.pyNN.utilities.constants import SPIKES
from spynnaker.pyNN.utilities.neo_csv import NeoCsv

if TYPE_CHECKING:
    from _csv import _writer as CSVWriter
    from spynnaker.pyNN.models.common.types import Names as ConcreteNames
    from spynnaker.pyNN.models.populations.population import Population
    from .data_population import DataPopulation
    #: :meta private:
    Names = Optional[ConcreteNames]
    #: :meta private:
    Annotations = Optional[Dict[str, Any]]

logger = FormatAdapter(logging.getLogger(__name__))

segment_cache: Dict[int, str] = {}


class NeoBufferDatabase(BufferDatabase, NeoCsv):
    """
    Extra support for Neo on top of the Database for SQLite 3.

    This is the same database as used by BufferManager but with
    extra tables and access methods added.
    """
    __N_BYTES_FOR_TIMESTAMP = BYTES_PER_WORD
    __TWO_WORDS = struct.Struct("<II")
    __NEO_DDL_FILE = os.path.join(os.path.dirname(__file__), "db.sql")
    #: rewiring: shift values to decode recorded value
    __PRE_ID_SHIFT = 9
    __POST_ID_SHIFT = 1
    __POST_ID_FACTOR = 2 ** 8
    __FIRST_BIT = 1
    #: number of words per rewiring entry
    __REWIRING_N_WORDS = 2

    @staticmethod
    def _string(value: _SqliteTypes) -> str:
        if isinstance(value, (bytes, memoryview)):
            return str(value, 'utf-8')
        else:
            return str(value)

    def __init__(self, database_file: Optional[str] = None,
                 read_only: Optional[bool] = None):
        """
        :param database_file:
            The name of a file that contains (or will contain) an SQLite
            database holding the data.
            If omitted the default location will be used.
        :param read_only:
            By default the database is read-only if given a database file.
            This allows to override that (mainly for clear)
        """
        if database_file is None:
            database_file = self.default_database_file()
            if read_only is None:
                read_only = False
        else:
            if read_only is None:
                read_only = True

        super().__init__(database_file, read_only=read_only)

        segment = SpynnakerDataView.get_reset_number()
        if (segment not in segment_cache or
                segment_cache[segment] != database_file):
            with open(self.__NEO_DDL_FILE, encoding="utf-8") as f:
                sql = f.read()
            # pylint: disable=no-member
            self._SQLiteDB__db.executescript(sql)  # type: ignore[attr-defined]
            segment_cache[segment] = database_file

    @classmethod
    def segement_db(cls, segment_number: int,
                    read_only: Optional[bool] = None) -> NeoBufferDatabase:
        """
        :returns: A NeoBufferDatabase for this segment.
        """
        database_file = segment_cache[segment_number]
        return NeoBufferDatabase(database_file, read_only)

    def write_segment_metadata(self) -> None:
        """
        Writes the global information from the Views.

        This writes information held in :py:class:`SpynnakerDataView` so that
        the database is usable stand-alone.

        .. note::
            The database must be writable for this to work!
        """
        # t_stop intentionally left None to show no run data
        self.cursor().execute(
            """
            INSERT INTO segment(
                simulation_time_step_ms, segment_number, rec_datetime,
                dt, simulator)
            VALUES (?, ?, ?, ?, ?)
            """, (SpynnakerDataView.get_simulation_time_step_ms(),
                  SpynnakerDataView.get_reset_number(),
                  datetime.now(),
                  SpynnakerDataView.get_simulation_time_step_ms(),
                  SpynnakerDataView.get_sim_name()))

    def write_t_stop(self) -> None:
        """
        Records the current run time as `t_Stop`.

        This writes information held in :py:class:`SpynnakerDataView` so that
        the database is usable stand-alone.

        .. note::
            The database must be writable for this to work!
        """
        t_stop = SpynnakerDataView.get_current_run_time_ms()
        self.cursor().execute(
            """
            UPDATE segment
            SET t_stop = ?
            """, (t_stop,))

    def __get_segment_info(self) -> Tuple[int, datetime, float, float, str]:
        """
        Gets the metadata for the segment.

        :return: segment number, record time, last run time recorded,
            simulator timestep in ms, simulator name
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording metadata not setup correctly
        """
        for row in self.cursor().execute(
                """
                SELECT segment_number, rec_datetime, t_stop, dt, simulator
                FROM segment
                LIMIT 1
                """):
            t_str = self._string(row[self._REC_DATETIME])
            time = datetime.strptime(t_str, "%Y-%m-%d %H:%M:%S.%f")
            if row[self._T_STOP] is None:
                t_stop = 0.0
                logger.warning("Data from a virtual run will be empty")
            else:
                t_stop = row[self._T_STOP]
            return (row[self._SEGMENT_NUMBER], time, t_stop, row[self._DT],
                    self._string(row[self._SIMULATOR]))
        raise ConfigurationException(
            "No recorded data. Did the simulation run?")

    def __get_simulation_time_step_ms(self) -> float:
        """
        The simulation time step, in milliseconds.

        The value that would be/have been returned by
        SpynnakerDataView.get_simulation_time_step_ms()

        :return: The timestep
        """
        for row in self.cursor().execute(
                """
                SELECT simulation_time_step_ms
                FROM segment
                LIMIT 1
                """):
            return row["simulation_time_step_ms"]
        raise ConfigurationException("No segment data")

    def __get_population_id(
            self, pop_label: str, population: Population) -> int:
        """
        Gets an ID for this population label.
        Will create a new population if required.

        For speed does not verify the additional fields if a record already
        exists.

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :param population: the population to record for
        :return: The ID
        """
        for row in self.cursor().execute(
                """
                SELECT pop_id FROM population
                WHERE label = ?
                LIMIT 1
                """, (pop_label,)):
            return row["pop_id"]
        self.cursor().execute(
            """
            INSERT INTO population(
                label, first_id, description, pop_size)
            VALUES (?, ?, ?, ?)
            """, (pop_label, population.first_id, population.describe(),
                  population.size))
        rowid = self.lastrowid
        assert rowid is not None, "population must have been inserted"
        return rowid

    def __get_recording_id(
            self, pop_label: str, variable: str,
            population: Population, sampling_interval_ms: Optional[float],
            data_type: Optional[DataType], buffered_type: BufferDataType,
            units: Optional[str], n_colour_bits: int) -> int:
        """
        Gets an ID for this population and recording label combination.
        Will create a new population/recording record if required.

        For speed does not verify the additional fields if a record already
        exists.

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :param variable:
        :param population: the population to record for
        :param sampling_interval_ms:
            The simulation time in milliseconds between sampling.
            Typically the sampling rate * simulation_timestep_ms
        :param buffered_type:
        :param units:
        :param n_colour_bits:
        :return: The ID
        """
        for row in self.cursor().execute(
                """
                SELECT rec_id FROM recording_view
                WHERE label = ? AND variable = ?
                LIMIT 1
                """, (pop_label, variable)):
            return row["rec_id"]
        pop_id = self.__get_population_id(pop_label, population)
        if data_type:
            data_type_name = data_type.name
        else:
            data_type_name = None
        self.cursor().execute(
            """
            INSERT INTO recording(
                pop_id, variable, data_type, buffered_type, t_start,
                sampling_interval_ms, units, n_colour_bits)
            VALUES (?, ?, ?, ?, 0, ?, ?, ?)
            """, (pop_id, variable, data_type_name, str(buffered_type),
                  sampling_interval_ms, units, n_colour_bits))
        rowid = self.lastrowid
        assert rowid is not None, "recording must have been inserted"
        return rowid

    def has_population_metadata(self, pop_label: str) -> bool:
        """
        Check if there is Metadata for the population with this label

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :returns: True if there is Metadata with this label
        """
        for _ in self.cursor().execute(
                """
                SELECT pop_size
                FROM population
                WHERE label = ?
                LIMIT 1
                """, (pop_label,)):
            return True
        return False

    def get_population_metadata(self, pop_label: str) -> Tuple[int, int, str]:
        """
        Gets the metadata for the population with this label

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :return: population size, first ID and description
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording metadata not setup correctly
        """
        for row in self.cursor().execute(
                """
                SELECT pop_size, first_id, description
                FROM population
                WHERE label = ?
                LIMIT 1
                """, (pop_label,)):
            return (int(row["pop_size"]), int(row["first_id"]),
                    self._string(row["description"]))
        raise ConfigurationException(f"There is no Metadata for {pop_label}")

    def get_recording_populations(self) -> Tuple[str, ...]:
        """
        Gets a list of the labels of Populations recording.
        Or to be exact the ones with metadata saved so likely to be recording.

        .. note::
            These are actually the labels of the Application Vertices.
            Typically the Population label, corrected for `None` or
            duplicate values

        :return: List of population labels
        """
        results = []
        for row in self.cursor().execute(
                """
                SELECT label
                FROM population
                """):
            results.append(self._string(row["label"]))
        return tuple(results)

    def get_population(self, pop_label: str) -> DataPopulation:
        """
        Gets an Object with the same data retrieval API as a Population.

        Retrieval is limited to recorded data and a little metadata needed to
        create a single Neo Segment wrapped in a Neo Block.

        .. note::
            As each database only includes data for one run (with resets
            creating another database) the structure is relatively simple.

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :return: An Object which acts like a Population for getting neo data
        """
        # delayed import due to circular dependencies
        # pylint: disable=import-outside-toplevel
        from .data_population import DataPopulation as DataPop
        # DataPopulation validates the pop_label so no need to do here too
        return DataPop(self._database_file, pop_label)

    def get_recording_variables(self, pop_label: str) -> Tuple[str, ...]:
        """
        List of the names of variables recording.
        Or, to be exact, list of the names of variables with metadata so likely
        to be recording.

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :return: List of variable names
        """
        results: List[str] = []
        for row in self.cursor().execute(
                """
                SELECT variable
                FROM recording_view
                WHERE label = ?
                GROUP BY variable
                """, (pop_label,)):
            results.append(self._string(row["variable"]))
        return tuple(results)

    def find_units(self, pop_label: str, variable: str) -> Optional[str]:
        """
        Gets the metadata ID for this population and recording label
        combination.

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :param variable:
        :return: data_type, sampling_interval_ms, units
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording metadata not setup correctly
        """
        info = self.__get_recording_metadata(pop_label, variable)
        if info is None:
            return None
        else:
            (_, _, _, _, _, _, units, _) = info
        return units

    def __get_recording_metadata(
            self, pop_label: str, variable: str) -> Optional[Tuple[
                int, Optional[DataType], BufferDataType, float, float, int,
                Optional[str], int]]:
        """
        Gets the metadata id for this population and recording label
        combination.

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typical the Population label, corrected for `None` or
                duplicate values

        :param variable:
        :return:
            id, data_type, buffered_type, t_start,
            sampling_interval_ms, pop_size, units, n_colour_bits
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording metadata not setup correctly
        """
        for row in self.cursor().execute(
                """
                SELECT rec_id,  data_type, buffered_type,  t_start,
                       sampling_interval_ms, pop_size, units, n_colour_bits
                FROM recording_view
                WHERE label = ? AND variable = ?
                LIMIT 1
                """, (pop_label, variable)):
            if row["data_type"]:
                data_type = DataType[self._string(row["data_type"])]
            else:
                data_type = None
            if row["units"]:
                units = self._string(row["units"])
            else:
                units = None
            buffered_type = BufferDataType[self._string(row["buffered_type"])]
            return (row["rec_id"], data_type, buffered_type, row["t_start"],
                    row["sampling_interval_ms"], row["pop_size"], units,
                    row["n_colour_bits"])
        return None

    def __get_region_metadata(self, rec_id: int) -> Iterable[Tuple[
            int, Optional[NDArray[integer]], Slice, Optional[bool], int, int]]:
        """
        :param rec_id:
        :return:
            region_id, neurons, vertex_slice, selective_recording, base_key,
            index
        """
        index = 0
        # Need to put the rows in a list to get them to persist.
        for row in list(self.cursor().execute(
                """
                SELECT region_id, recording_neurons_st, vertex_slice, base_key
                FROM region_metadata
                WHERE rec_id = ?
                ORDER BY region_id, recording_neurons_st, vertex_slice,
                    base_key
                """, (rec_id,))):
            vertex_slice = MDSlice.from_string(
                self._string(row["vertex_slice"]))
            recording_neurons_st = row["recording_neurons_st"]
            if recording_neurons_st:
                neurons = numpy.array(
                    self.string_to_array(recording_neurons_st))
                yield (row["region_id"], neurons, vertex_slice,
                       len(neurons) != vertex_slice.n_atoms,
                       row["base_key"], index)
            else:
                yield (row["region_id"], None, vertex_slice, None,
                       row["base_key"], index)
            index += 1

    def __get_spikes_by_region(
            self, region_id: int, neurons: NDArray[integer],
            simulation_time_step_ms: float, selective_recording: bool,
            spike_times: List[float], spike_ids: List[int]) -> None:
        """
        Adds spike data for this region to the lists.

        :param region_id: Region data came from
        :param neurons: mapping of local ID to global ID
        :param simulation_time_step_ms:
        :param selective_recording: flag to say if
        :param spike_times: List to add spike times to
        :param spike_ids: List to add spike IDs to
        """
        neurons_recording = len(neurons)
        if neurons_recording == 0:
            return
        n_words = int(math.ceil(neurons_recording / BITS_PER_WORD))
        n_bytes = n_words * BYTES_PER_WORD
        n_words_with_timestamp = n_words + 1

        record_raw = self._read_recording(region_id)

        if len(record_raw) == 0:
            return

        raw_data = (
            numpy.asarray(record_raw, dtype=uint8).view(
                dtype="<i4")).reshape([-1, n_words_with_timestamp])

        record_time = (raw_data[:, 0] * simulation_time_step_ms)
        spikes = raw_data[:, 1:].byteswap().view(uint8)
        bits = numpy.fliplr(numpy.unpackbits(spikes).reshape(
            (-1, 32))).reshape((-1, n_bytes * 8))
        time_indices, local_indices = numpy.where(bits == 1)
        if selective_recording:
            for time_indice, local in zip(time_indices, local_indices):
                if local < neurons_recording:
                    spike_ids.append(neurons[local])
                    spike_times.append(record_time[time_indice])
        else:
            indices = neurons[local_indices]
            times = record_time[time_indices].reshape((-1))
            spike_ids.extend(indices)
            spike_times.extend(times)

    def __get_neuron_spikes(self, rec_id: int) -> Tuple[
            NDArray, List[int]]:
        """
        Gets the spikes for this population/recording ID.

        :param rec_id:
        :return: numpy array of spike IDs and spike times, all IDs recording
        """
        spike_times: List[float] = []
        spike_ids: List[int] = []
        simulation_time_step_ms = self.__get_simulation_time_step_ms()
        indexes: List[int] = []
        for region_id, neurons, _, selective_recording, _, _ in \
                self.__get_region_metadata(rec_id):
            if neurons is None or selective_recording is None:
                continue
            indexes.extend(neurons)
            self.__get_spikes_by_region(
                region_id, neurons, simulation_time_step_ms,
                selective_recording, spike_times, spike_ids)

        result = numpy.column_stack((spike_ids, spike_times))
        return result[numpy.lexsort(result.T[::-1])], indexes

    def __get_eieio_spike_by_region(
            self, region_id: int,
            simulation_time_step_ms: float, base_key: int,
            vertex_slice: Slice, n_colour_bits: int,
            results: List[NDArray]) -> NDArray[integer]:
        """
        Adds spike data for this region to the list.

        :param region_id: Region data came from
        :param simulation_time_step_ms:
        :param base_key:
        :param vertex_slice:
        :param n_colour_bits:
        :param results: Where to add spike data to
        :return: all recording indexes spikes or not
        """
        spike_data = self._read_recording(region_id)

        number_of_bytes_written = len(spike_data)
        offset = 0
        indices = {
            key: index for index, key in enumerate(
                get_keys(base_key, vertex_slice, n_colour_bits))}
        slice_ids = vertex_slice.get_raster_ids()
        colour_mask = (2 ** n_colour_bits) - 1
        inv_colour_mask = ~colour_mask & 0xFFFFFFFF
        while offset < number_of_bytes_written:
            length, time = self.__TWO_WORDS.unpack_from(spike_data, offset)
            time *= simulation_time_step_ms
            data_offset = offset + 2 * BYTES_PER_WORD

            eieio_header = EIEIODataHeader.from_bytestring(
                spike_data, data_offset)
            if eieio_header.eieio_type.payload_bytes > 0:
                raise ValueError("Can only read spikes as keys")

            data_offset += eieio_header.size
            timestamps = numpy.repeat([time], eieio_header.count)
            key_bytes = eieio_header.eieio_type.key_bytes
            keys = numpy.frombuffer(
                spike_data, dtype=f"<u{key_bytes}",
                count=eieio_header.count, offset=data_offset).astype(uint32)
            keys = numpy.bitwise_and(keys, inv_colour_mask)
            local_ids = numpy.array([indices[key] for key in keys])
            neuron_ids = slice_ids[local_ids]
            offset += length + 2 * BYTES_PER_WORD
            results.append(numpy.dstack((neuron_ids, timestamps))[0])

        return slice_ids

    def __get_eieio_spikes(
            self, rec_id: int, n_colour_bits: int) -> Tuple[
                NDArray, List[int]]:
        """
        Gets the spikes for this population/recording ID.

        :return: numpy array of spike IDs and spike times, all IDs recording
        """
        simulation_time_step_ms = self.__get_simulation_time_step_ms()
        results: List[NDArray] = []
        indexes: List[int] = []

        for region_id, _, vertex_slice, selective_recording, base_key, _ in \
                self.__get_region_metadata(rec_id):
            if selective_recording:
                raise NotImplementedError(
                    "Unable to handle selective recording")
            indexes.extend(self.__get_eieio_spike_by_region(
                region_id, simulation_time_step_ms,
                base_key, vertex_slice, n_colour_bits, results))

        if not results:
            return numpy.empty(shape=(0, 2)), indexes
        result = numpy.vstack(results)
        return result[numpy.lexsort((result[:, 1], result[:, 0]))], indexes

    def __get_multi_spikes_by_region(
            self, region_id: int, neurons: NDArray[integer],
            simulation_time_step_ms: float,
            spike_times: List[NDArray[floating]],
            spike_ids: List[NDArray[integer]]) -> None:
        """
        Adds spike data for this region to the lists.

        :param region_id: Region data came from
        :param neurons:
        :param simulation_time_step_ms:
        :param spike_times: List to add spike times to
        :param spike_ids: List to add spike IDs to
        """
        raw_data = self._read_recording(region_id)

        n_words = int(math.ceil(len(neurons) / BITS_PER_WORD))
        n_bytes_per_block = n_words * BYTES_PER_WORD
        offset = 0
        while offset < len(raw_data):
            time, n_blocks = self.__TWO_WORDS.unpack_from(raw_data, offset)
            offset += self.__TWO_WORDS.size
            spike_data = numpy.frombuffer(
                raw_data, dtype=uint8,
                count=n_bytes_per_block * n_blocks, offset=offset)
            offset += n_bytes_per_block * n_blocks

            spikes = spike_data.view("<i4").byteswap().view(uint8)
            bits = numpy.fliplr(numpy.unpackbits(spikes).reshape(
                (-1, 32))).reshape((-1, n_bytes_per_block * 8))
            local_indices = numpy.nonzero(bits)[1]
            indices = neurons[local_indices]
            times = numpy.repeat(
                [time * simulation_time_step_ms],
                len(indices))
            spike_ids.append(indices)
            spike_times.append(times)

    def __get_multi_spikes(self, rec_id: int) -> Tuple[
            NDArray, List[int]]:
        """
        Gets the spikes for this population/recording ID.

        :param rec_id:
        :return: numpy array of spike IDs and spike times, all IDs recording
        """
        spike_times_l: List[NDArray[floating]] = []
        spike_ids_l: List[NDArray[integer]] = []
        indexes: List[int] = []
        simulation_time_step_ms = self.__get_simulation_time_step_ms()
        for region_id, neurons, _, selective_recording, _, _ in \
                self.__get_region_metadata(rec_id):
            if selective_recording:
                raise NotImplementedError(
                    "Unable to handle selective recording")
            assert neurons is not None
            indexes.extend(neurons)
            self.__get_multi_spikes_by_region(
                region_id, neurons, simulation_time_step_ms,
                spike_times_l, spike_ids_l)

        if not spike_ids_l:
            return numpy.zeros((0, 2)), indexes

        spike_ids_a = numpy.hstack(spike_ids_l)
        spike_times_a = numpy.hstack(spike_times_l)
        result = numpy.dstack((spike_ids_a, spike_times_a))[0]
        return result[numpy.lexsort((spike_times_a, spike_ids_a))], indexes

    @staticmethod
    def __combine_indexes(
            view_indexes: Union[Sequence[int], NDArray[integer]],
            data_indexes: Union[Sequence[int], NDArray[integer]],
            variable: str) -> NDArray[integer]:
        """
        :param view_indexes:
        :param data_indexes:
        :param variable:
        :return: indices
        """
        # keep just the view indexes in the data
        data_set = set(data_indexes)
        indexes = [i for i in view_indexes if i in data_set]
        # check for missing and report
        view_set: Set[int] = set(view_indexes)
        missing = view_set.difference(data_indexes)
        if missing:
            logger.warning("No {} available for neurons {}",
                           variable, sorted(missing))
        return numpy.array(indexes)

    def __get_spikes(
            self, rec_id: int, view_indexes: ViewIndices,
            buffer_type: BufferDataType, n_colour_bits: int,
            variable: str) -> Tuple[NDArray, NDArray[integer]]:
        """
        Gets the data as a Numpy array for one population and variable.

        :param rec_id:
        :param view_indexes:
        :param buffer_type:
        :param n_colour_bits:
        :param variable:
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording metadata not setup correctly
        """
        if buffer_type == BufferDataType.NEURON_SPIKES:
            spikes, data_indexes = self.__get_neuron_spikes(rec_id)
        elif buffer_type == BufferDataType.EIEIO_SPIKES:
            spikes, data_indexes = self.__get_eieio_spikes(
                rec_id, n_colour_bits)
        elif buffer_type == BufferDataType.MULTI_SPIKES:
            spikes, data_indexes = self.__get_multi_spikes(rec_id)
        else:
            raise NotImplementedError(buffer_type)

        if view_indexes is None or list(view_indexes) == list(data_indexes):
            indexes = numpy.array(data_indexes)
        else:
            # keep just the view indexes in the data
            indexes = self.__combine_indexes(
                view_indexes, data_indexes, variable)
            # keep just data columns in the view
            spikes = spikes[numpy.isin(spikes[:, 0], indexes)]

        return spikes, indexes

    def __get_matrix_data_by_region(
            self, region_id: int, neurons: NDArray[integer],
            data_type: DataType) -> Tuple[
                NDArray[floating], NDArray[floating]]:
        """
        Extracts data for this region.

        :param region_id: Region data came from
        :param neurons: mapping of local ID to global ID
        :param data_type: type of data to extract
        :return: times, data
        """
        # for buffering output info is taken form the buffer manager
        record_raw = self._read_recording(region_id)
        record_length = len(record_raw)

        # There is one column for time and one for each neuron recording
        data_row_length = len(neurons) * data_type.size
        full_row_length = data_row_length + self.__N_BYTES_FOR_TIMESTAMP
        n_rows = record_length // full_row_length
        row_data = numpy.asarray(record_raw, dtype=uint8).reshape(
            n_rows, full_row_length)

        time_bytes = (
            row_data[:, 0: self.__N_BYTES_FOR_TIMESTAMP].reshape(
                n_rows * self.__N_BYTES_FOR_TIMESTAMP))
        times = time_bytes.view("<i4").reshape(n_rows, 1)
        var_data = (row_data[:, self.__N_BYTES_FOR_TIMESTAMP:].reshape(
            n_rows * data_row_length))
        placement_data = data_type.decode_array(var_data).reshape(
            n_rows, len(neurons))

        return times, placement_data

    def __get_matrix_data(
            self, rec_id: int, data_type: DataType,
            view_indexes: ViewIndices, pop_size: int,
            variable: str) -> Tuple[NDArray[floating], NDArray[integer]]:
        """
        Gets the matrix data  for this population/recording ID.

        :param rec_id:
        :param data_type: type of data to extract
        :param view_indexes:
            The indexes for which data should be returned. Or `None` for all
        :param pop_size:
        :param variable:
        :return: numpy array of the data, neurons
        """
        signal_array: Optional[NDArray[floating]] = None
        pop_times: Optional[NDArray[floating]] = None
        pop_neurons: List[None] = []
        indexes: List[int] = []

        for region_id, neurons, _, _, _, index in \
                self.__get_region_metadata(rec_id):
            if neurons is not None:
                pop_neurons.extend(neurons)
            else:
                indexes.append(index)
                neurons = numpy.array([index], dtype=uint32)
            times, data = self.__get_matrix_data_by_region(
                region_id, neurons, data_type)
            if signal_array is None or pop_times is None:
                signal_array = data
                pop_times = times
            elif numpy.array_equal(pop_times, times):
                signal_array = numpy.append(
                    signal_array, data, axis=1)
            else:
                raise NotImplementedError("times differ")
        if signal_array is None:
            signal_array = numpy.zeros((0,), dtype=float64)

        if len(indexes) > 0:
            assert (len(pop_neurons) == 0)
            if view_indexes is not None:
                raise SpynnakerException(
                    f"{variable} data can not be extracted using a view")
            return signal_array, numpy.array(indexes)

        data_indexes = numpy.array(pop_neurons)
        if view_indexes is None:
            view_indexes = range(pop_size)
        if list(view_indexes) == list(data_indexes):
            indexes_a = numpy.array(data_indexes)
        else:
            # keep just the view indexes in the data
            indexes_a = self.__combine_indexes(
                view_indexes, data_indexes, variable)
            # keep just data columns in the view
            map_indexes = [list(data_indexes).index(i) for i in indexes_a]
            signal_array = signal_array[:, map_indexes]

        return signal_array, indexes_a

    def __get_rewires_by_region(
            self, region_id: int, vertex_slice: Slice,
            rewire_values: List[int], rewire_postids: List[int],
            rewire_preids: List[int], rewire_times: List[int],
            sampling_interval_ms: float) -> None:
        """
        Extracts rewires data for this region and adds it to the lists.

        :param region_id: Region data came from
        :param vertex_slice: slice of this region
        :param rewire_values:
        :param rewire_postids:
        :param rewire_preids:
        :param rewire_times:
        """
        record_raw = self._read_recording(region_id)
        if len(record_raw) < 1:
            return

        raw_data = numpy.asarray(record_raw, dtype=uint8).view(
            dtype="<i4").reshape([-1, self.__REWIRING_N_WORDS])

        record_time = (raw_data[:, 0] * sampling_interval_ms)
        rewires_raw = raw_data[:, 1:]
        rew_length = len(rewires_raw)
        # rewires is 0 (elimination) or 1 (formation) in the first bit
        rewires = (rewires_raw[i][0] & self.__FIRST_BIT
                   for i in range(rew_length))
        # the post-neuron ID is stored in the next 8 bytes
        post_ids = (((int(rewires_raw[i]) >> self.__POST_ID_SHIFT) %
                     self.__POST_ID_FACTOR) + vertex_slice.lo_atom
                    for i in range(rew_length))
        # the pre-neuron ID is stored in the remaining 23 bytes
        pre_ids = (int(rewires_raw[i]) >> self.__PRE_ID_SHIFT
                   for i in range(rew_length))

        rewire_values.extend(rewires)
        rewire_postids.extend(post_ids)
        rewire_preids.extend(pre_ids)
        rewire_times.extend(record_time)

    def __get_rewires(self, rec_id: int,
                      sampling_interval_ms: float) -> NDArray[integer]:
        """
        Extracts rewires data for this region.

        :param rec_id:
        :return: (rewire_values, rewire_postids, rewire_preids, rewire_times)
        """
        rewire_times: List[int] = []
        rewire_values: List[int] = []
        rewire_postids: List[int] = []
        rewire_preids: List[int] = []

        for region_id, _, vertex_slice, _, _, _ in \
                self.__get_region_metadata(rec_id):
            # as no neurons for "rewires" selective_recording will be true

            self.__get_rewires_by_region(
                region_id, vertex_slice, rewire_values, rewire_postids,
                rewire_preids, rewire_times, sampling_interval_ms)

            if len(rewire_values) == 0:
                return numpy.zeros((0, 4), dtype=uint32)

        result = numpy.column_stack(
            (rewire_times, rewire_preids, rewire_postids, rewire_values))
        return result[numpy.lexsort(
            (rewire_values, rewire_postids, rewire_preids, rewire_times))]

    def __get_recorded_pynn7(
            self, rec_id: int, data_type: DataType,
            sampling_interval_ms: float, as_matrix: bool,
            view_indexes: ViewIndices, pop_size: int,
            variable: str) -> NDArray[floating]:
        """
        Get recorded data in PyNN 0.7 format. Must not be spikes.

        :param rec_id:
        :param data_type: type of data to extract
        :param sampling_interval_ms:
        :param as_matrix:
        :param view_indexes:
            The indexes for which data should be returned. Or `None` for all
        :param pop_size:
        :param variable:
        """
        data, indexes = self.__get_matrix_data(
            rec_id, data_type, view_indexes, pop_size, variable)

        if as_matrix:
            return data

        # Convert to triples as Pynn 0,7 did
        n_machine_time_steps = len(data)
        n_neurons = len(indexes)
        column_length = n_machine_time_steps * n_neurons
        times = [i * sampling_interval_ms
                 for i in range(0, n_machine_time_steps)]
        return numpy.column_stack((
                numpy.repeat(indexes, n_machine_time_steps, 0),
                numpy.tile(times, n_neurons),
                numpy.transpose(data).reshape(column_length)))

    def spinnaker_get_data(
            self, pop_label: str, variable: str, as_matrix: bool = False,
            view_indexes: ViewIndices = None) -> NDArray[floating]:
        """
        SsPyNNaker specific method for getting data as a numpy array,
        instead of the Neo-based object

        :param pop_label: label for the Population
        :param variable: Single variable name.
        :param as_matrix: If set True the data is returned as a 2d matrix
        :param view_indexes: The indexes for which data should be returned.
            If ``None``, all data (view_index = data_indexes)
        :return: array of the data
        :raises ConfigurationException:
            If variable is a list of a length other than 1
        """
        if not isinstance(variable, str):
            if len(variable) != 1:
                raise ConfigurationException(
                    "Only one type of data at a time is supported")
            variable = variable[0]

        # called to trigger the virtual data warning if applicable
        self.__get_segment_info()
        metadata = self.__get_recording_metadata(pop_label, variable)
        if metadata is None:
            return numpy.empty((0, 0), dtype=float)
        (rec_id, data_type, buffered_type, _, sampling_interval_ms,
         pop_size, _, n_colour_bits) = metadata

        if buffered_type == BufferDataType.MATRIX:
            assert data_type is not None
            return self.__get_recorded_pynn7(
                rec_id, data_type, sampling_interval_ms,
                as_matrix, view_indexes, pop_size, variable)
        # NO BufferedDataType.REWIRES get_spike will go boom
        else:
            if as_matrix:
                logger.warning("Ignoring as matrix for {}", variable)
            return self.__get_spikes(
                rec_id, view_indexes, buffered_type,
                n_colour_bits, variable)[0]

    def get_spike_counts(
            self, pop_label: str,
            view_indexes: ViewIndices = None) -> Dict[int, int]:
        """
        Gets the spike counts for the population with this label.

        :param pop_label: label for the Population
        :param view_indexes: If supplied indexes to retrieve.
        :return: dict of index to count
        """
        # called to trigger the virtual data warning if applicable
        self.__get_segment_info()
        metadata = self.__get_recording_metadata(pop_label, SPIKES)
        if metadata is None:
            raise ConfigurationException(
                f"{pop_label} did not record spikes")

        (rec_id, _, buffered_type, _, _, pop_size, _,
         n_colour_bits) = metadata

        if view_indexes is None:
            view_indexes = range(pop_size)

        # get_spike will go boom if buffered_type not spikes
        spikes = self.__get_spikes(
            rec_id, view_indexes, buffered_type,
            n_colour_bits, SPIKES)[0]
        counts = numpy.bincount(spikes[:, 0].astype(dtype=numpy.int32),
                                minlength=pop_size)
        return {i: counts[i] for i in view_indexes}

    def __add_data(
            self, pop_label: str, variable: str,
            segment: neo.Segment, view_indexes: ViewIndices, t_stop: float,
            allow_missing: bool) -> None:
        """
        Gets the data as a Numpy array for one population and variable.

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :param variable:
        :param segment: Segment to add data to
        :param t_stop:
        :param allow_missing: If True silently skips is variable not recorded
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording metadata not setup correctly
        """
        metadata = self.__get_recording_metadata(pop_label, variable)
        if metadata is None:
            if allow_missing:
                return
            else:
                raise ConfigurationException(
                    f"No data for {pop_label=} {variable=}")

        (rec_id, data_type, buffer_type, t_start, sampling_interval_ms,
         pop_size, units, n_colour_bits) = metadata

        if buffer_type == BufferDataType.MATRIX:
            assert data_type is not None
            signal_array, indexes = self.__get_matrix_data(
                rec_id, data_type, view_indexes, pop_size, variable)
            sampling_rate = 1000/sampling_interval_ms * quantities.Hz
            t_start = t_start * quantities.ms
            self._insert_matrix_data(
                variable, segment, signal_array,
                indexes, t_start, sampling_rate, units)
        elif buffer_type == BufferDataType.REWIRES:
            if view_indexes is not None:
                raise SpynnakerException(
                    f"{variable} can not be extracted using a view")
            event_array = self.__get_rewires(
                rec_id, sampling_interval_ms)
            self._insert_neo_rewirings(segment, event_array, variable)
        else:
            if view_indexes is None:
                view_indexes = range(pop_size)
            spikes, indexes = self.__get_spikes(
                rec_id, view_indexes, buffer_type,
                n_colour_bits, variable)
            sampling_rate = 1000 / sampling_interval_ms * quantities.Hz
            self._insert_spike_data(
                view_indexes, segment, spikes, t_start, t_stop,
                sampling_rate)

    def __read_and_csv_data(
            self, pop_label: str, variable: str, csv_writer: CSVWriter,
            view_indexes: ViewIndices, t_stop: float,
            allow_missing: bool) -> None:
        """
        Reads the data for one variable and adds it to the CSV file.

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :param variable:
        :param csv_writer: Open CSV writer to write to
        :param view_indexes:
        :param t_stop:
        :param allow_missing: Flag to say if data for missing variable
            should raise an exception
        """
        metadata = self.__get_recording_metadata(pop_label, variable)
        if metadata is None:
            if allow_missing:
                return
            else:
                raise ConfigurationException(
                    f"No data for {pop_label=} {variable=}")

        (rec_id, data_type, buffer_type, t_start, sampling_interval_ms,
         pop_size, units, n_colour_bits) = metadata

        if buffer_type == BufferDataType.MATRIX:
            assert data_type is not None
            self._csv_variable_metdata(
                csv_writer, self._MATRIX, variable, t_start, t_stop,
                sampling_interval_ms, units)
            signal_array, indexes = self.__get_matrix_data(
                rec_id, data_type, view_indexes, pop_size, variable)
            self._csv_matrix_data(csv_writer, signal_array, indexes)
        elif buffer_type == BufferDataType.REWIRES:
            self._csv_variable_metdata(
                csv_writer, self._EVENT, variable, t_start, t_stop,
                sampling_interval_ms, units)
            if view_indexes is not None:
                raise SpynnakerException(
                    f"{variable} can not be extracted using a view")
            event_array = self.__get_rewires(rec_id, sampling_interval_ms)
            self._csv_rewirings(csv_writer, event_array)
        else:
            self._csv_variable_metdata(
                csv_writer, self._SPIKES, variable, t_start, t_stop,
                sampling_interval_ms, units)
            spikes, indexes = self.__get_spikes(
                rec_id, view_indexes, buffer_type, n_colour_bits, variable)
            self._csv_spike_data(csv_writer, spikes, indexes)

    def get_empty_block(self, pop_label: str,
                        annotations: Annotations) -> Optional[neo.Block]:
        """
        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :param annotations: annotations to put on the neo block
        :return: The Neo block
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording metadata not setup correctly
        """
        _, _, _, dt, simulator = self.__get_segment_info()
        if self.has_population_metadata(pop_label):
            metadata = self.get_population_metadata(pop_label)
        else:
            return None
        pop_size, first_id, description = metadata

        return self._insert_empty_block(
            pop_label, description, pop_size, first_id, dt, simulator,
            annotations)

    def get_full_block(
            self, pop_label: str, variables: Names, view_indexes: ViewIndices,
            annotations: Annotations) -> neo.Block:
        """
        Creates a block with metadata and data for this segment.
        Any previous segments will be empty.

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :param variables:
            One or more variable names or `None` for all available
        :param view_indexes: List of neurons IDs to include or `None` for all
        :param annotations: annotations to put on the neo block
        :return: The Neo block
        """
        block = self.get_empty_block(pop_label, annotations)
        self.add_segment(block, pop_label, variables, view_indexes,
                         allow_missing=False)
        return block

    def csv_segment(
            self, csv_file: str, pop_label: str, variables: Names,
            view_indexes: ViewIndices, allow_missing: bool) -> None:
        """
        Writes the data including metadata to a CSV file.

        :param csv_file: Path to file to write block metadata to
        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typical the Population label, corrected for `None` or
                duplicate values

        :param variables:
            One or more variable names or `None` for all available
        :param view_indexes: List of neurons IDs to include or `None` for all
        :param allow_missing: Flag to say if data for missing variable
            should raise an exception
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording metadata not setup correctly
        """
        if not os.path.isfile(csv_file):
            raise SpynnakerException("Please call csv_block_metadata first")
        with open(csv_file, 'a', newline='', encoding="utf-8") as csvfile:
            csv_writer = csv.writer(csvfile, delimiter=',', quotechar='"',
                                    quoting=csv.QUOTE_MINIMAL)

            segment_number, rec_datetime, t_stop, _, _ = \
                self.__get_segment_info()
            self._csv_segment_metadata(
                csv_writer, segment_number, rec_datetime)

            for variable in self.__clean_variables(variables, pop_label):
                self.__read_and_csv_data(pop_label, variable, csv_writer,
                                         view_indexes, t_stop, allow_missing)

    def csv_block_metadata(
            self, csv_file: str, pop_label: str,
            annotations: Annotations = None) -> bool:
        """
        Writes the data including metadata to a CSV file.
        Overwrites any previous data in the file.

        :param csv_file: Path to file to write block metadata to
        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :param annotations: annotations to put on the neo block
        :returns: True if metadata was available and therefore written
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording metadata not setup correctly
        """
        _, _, _, dt, _ = self.__get_segment_info()
        if self.has_population_metadata(pop_label):
            metadata = self.get_population_metadata(pop_label)
        else:
            return False

        pop_size, first_id, description = metadata
        with open(csv_file, 'w', newline='',  encoding="utf-8") as csvfile:
            csv_writer = csv.writer(csvfile, delimiter=',', quotechar='"',
                                    quoting=csv.QUOTE_MINIMAL)

            self._csv_block_metadata(
                csv_writer, pop_label, dt, pop_size, first_id, description,
                annotations)
        return True

    def __clean_variables(
            self, variables: Names, pop_label: str) -> Tuple[str, ...]:
        if variables is None:
            vs: Tuple[str, ...] = ("all", )
        elif isinstance(variables, str):
            vs = (variables, )
        else:
            vs = tuple(variables)
        if 'all' in vs:
            return self.get_recording_variables(pop_label)
        else:
            return vs

    def add_segment(
            self, block: neo.Block, pop_label: str, variables: Names,
            view_indexes: ViewIndices, allow_missing: bool) -> None:
        """
        Adds a segment to the block.

        :param  block:
        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typically the Population label, corrected for `None` or
                duplicate values

        :param variables:
            One or more variable names or `None` for all available
        :param view_indexes: List of neurons IDs to include or `None` for all
        :param allow_missing: If True silently skips any variables not recorded
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording metadata not setup correctly
        """
        segment_number, rec_datetime, t_stop, _, _ = \
            self.__get_segment_info()
        segment = self._insert_empty_segment(
            block, segment_number, rec_datetime)

        for variable in self.__clean_variables(variables, pop_label):
            self.__add_data(pop_label, variable, segment, view_indexes,
                            t_stop, allow_missing)

    def clear_data(self, pop_label: str, variables: Names) -> None:
        """
        Clears the data for one population and given variables.

        .. note:::
            The database must be writable for this to work!

        :param pop_label: The label for the population of interest

            .. note::
                This is actually the label of the Application Vertex.
                Typical the Population label, corrected for `None` or
                duplicate values

        :param variables: names of variable to get data for
        """
        t_start = SpynnakerDataView.get_current_run_time_ms()
        variables = self.__clean_variables(variables, pop_label)
        region_ids = []
        for variable in variables:
            self.cursor().execute(
                """
                UPDATE recording SET
                    t_start = ?
                WHERE rec_id in (
                    SELECT rec_id
                    FROM recording_view
                    WHERE label = ? AND variable = ?)
                """, (t_start, pop_label, variable))

            for row in self.cursor().execute(
               """
                    SELECT region_id
                    FROM region_metadata NATURAL JOIN recording_view
                    WHERE label = ? AND variable = ?
                    """, (pop_label, variable)):
                region_ids.append(int((row["region_id"])))

        for region_id in region_ids:
            self._clear_recording_region(region_id)

    def write_metadata(self) -> None:
        """
        Write the current metadata to the database.

        The underlying call does not guarantee order
        so there order the metadata is added is not consistent,

        .. note::
            The database must be writable for this to work!
        """
        for population in SpynnakerDataView.iterate_populations():
            # pylint: disable=protected-access
            for variable in population._vertex.get_recording_variables():
                self.__write_metadata(population, variable)

    def __write_metadata(self, population: Population, variable: str) -> None:
        # pylint: disable=protected-access
        app_vertex = population._vertex
        assert app_vertex.label is not None
        buffered_data_type = \
            app_vertex.get_buffer_data_type(variable)
        data_type = app_vertex.get_data_type(variable)
        sampling_interval_ms = \
            app_vertex.get_sampling_interval_ms(variable)

        units = app_vertex.get_units(variable)
        n_colour_bits = app_vertex.n_colour_bits
        rec_id = self.__get_recording_id(
            app_vertex.label, variable,
            population, sampling_interval_ms, data_type,
            buffered_data_type, units, n_colour_bits)
        region = app_vertex.get_recording_region(variable)
        machine_vertices = app_vertex.splitter.machine_vertices_for_recording(
            variable)
        for vertex in machine_vertices:
            placement = SpynnakerDataView.get_placement_of_vertex(
                vertex)
            region_id = self._get_recording_region_id(
                placement.x, placement.y, placement.p, region)
            vertex_slice = vertex.vertex_slice
            neurons = app_vertex.get_neurons_recording(
                variable, vertex_slice)
            if neurons is None:
                recording_neurons_st = None
            elif len(neurons) == 0:
                continue
            else:
                recording_neurons_st = self.array_to_string(neurons)
            if buffered_data_type == BufferDataType.EIEIO_SPIKES:
                # Sneaky! An undeclared interface...
                assert isinstance(
                    vertex, ReverseIPTagMulticastSourceMachineVertex)
                base_key = vertex.get_virtual_key()
            else:
                base_key = None
            self.cursor().execute(
                """
                INSERT INTO region_metadata(
                    rec_id, region_id, recording_neurons_st,
                    base_key, vertex_slice)
                VALUES (?, ?, ?, ?, ?)
                """,
                (rec_id, region_id, recording_neurons_st,
                 base_key, str(vertex.vertex_slice)))

    @staticmethod
    def array_to_string(indexes: Collection[int]) -> str:
        """
        Converts a list of non-negative integers into a compact string.
        Works best if the list is sorted.

        IDs are comma separated, except when a series of IDs is sequential then
        the start:end is used.

        :param indexes: Collection (ideally sorted of int Values)
        :returns: string representation to be used in the database
        """
        if indexes is None or len(indexes) == 0:
            return ""

        previous = -1
        results = ""
        in_range = False
        for index in indexes:
            if previous < 0:
                previous = index
                results = str(previous)
            else:
                if index == previous + 1:
                    if not in_range:
                        results += ":"
                        in_range = True
                else:
                    if in_range:
                        results += str(previous)
                    results += ","
                    results += str(index)
                    in_range = False
                previous = index
        if in_range:
            results += str(previous)
        return results

    @classmethod
    def string_to_array(cls, string: str) -> List[int]:
        """
        Converts a string into a list of integers.
        Assumes the string was created by :py:meth:`array_to_string`

        :param string: in format used by array_to_string
        :returns: List of integers
        """
        if not string:
            return []
        string = cls._string(string)
        results: List[int] = []
        parts = re.findall(r"\d+[,:]*", string)
        start = None
        for part in parts:
            if part.endswith(":"):
                start = int(part[:-1])
            else:
                if part.endswith(","):
                    val = int(part[:-1])
                else:
                    val = int(part)
                if start is not None:
                    results.extend(range(start, val+1))
                    start = None
                else:
                    results.append(val)

        return results



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple
import numpy
from numpy import integer, uint32
from numpy.typing import NDArray


def _calculate_stdp_times(
        pre_spikes: NDArray[integer], post_spikes: NDArray[integer],
        plastic_delay: int) -> Tuple[NDArray[integer], NDArray[integer]]:
    # If no post spikes, no changes
    if len(post_spikes) == 0:
        return numpy.zeros(0, dtype=uint32), numpy.zeros(0, dtype=uint32)

    # Get the spikes and time differences that will be considered by
    # the simulation (as the last pre-spike will be considered differently)
    last_pre_spike_delayed = pre_spikes[-1] - plastic_delay
    considered_post_spikes = post_spikes[post_spikes < last_pre_spike_delayed]
    if len(considered_post_spikes) == 0:
        return numpy.zeros(0, dtype=uint32), numpy.zeros(0, dtype=uint32)
    potentiation_time_diff = numpy.ravel(numpy.subtract.outer(
        considered_post_spikes + plastic_delay, pre_spikes[:-1]))
    potentiation_times = (
        potentiation_time_diff[potentiation_time_diff > 0] * -1)
    depression_time_diff = numpy.ravel(numpy.subtract.outer(
        considered_post_spikes + plastic_delay, pre_spikes))
    depression_times = depression_time_diff[depression_time_diff < 0]
    return potentiation_times, depression_times


def calculate_spike_pair_additive_stdp_weight(
        pre_spikes: NDArray[integer], post_spikes: NDArray[integer],
        initial_weight: float, plastic_delay: int,
        a_plus: float, a_minus: float,
        tau_plus: float, tau_minus: float) -> float:
    """
    Calculates the expected STDP weight for SpikePair Additive STDP.

    :param pre_spikes:
    :param post_spikes:
    :param initial_weight:
    :param plastic_delay: parameter of the STDP model
    :param a_plus: parameter of the STDP model
    :param a_minus: parameter of the STDP model
    :param tau_plus: parameter of the STDP model
    :param tau_minus: parameter of the STDP model
    :return: overall weight
    """
    potentiation_times, depression_times = _calculate_stdp_times(
        pre_spikes, post_spikes, plastic_delay)

    # Work out the weight according to the additive rule
    potentiations = a_plus * numpy.exp(potentiation_times / tau_plus)
    depressions = a_minus * numpy.exp(depression_times / tau_minus)

    # print("Potentiations: ", potentiation_times, potentiations)
    # print("Depressions:", depression_times, depressions)
    return initial_weight + numpy.sum(potentiations) - numpy.sum(depressions)


def calculate_spike_pair_multiplicative_stdp_weight(
        pre_spikes: NDArray[integer], post_spikes: NDArray[integer],
        initial_weight: float, plastic_delay: int,
        min_weight: float, max_weight: float,
        a_plus: float, a_minus: float,
        tau_plus: float, tau_minus: float) -> float:
    """
    Calculates the expected STDP weight for SpikePair Multiplicative STDP.

    :param pre_spikes: Spikes going into the model
    :param post_spikes: Spikes recorded on the model
    :param initial_weight: Starting weight for the model
    :param plastic_delay: parameter of the STDP model
    :param min_weight: parameter of the STDP model
    :param max_weight: parameter of the STDP model
    :param a_plus: parameter of the STDP model
    :param a_minus: parameter of the STDP model
    :param tau_plus: parameter of the STDP model
    :param tau_minus: parameter of the STDP model
    :return: overall weight
    """
    potentiation_times, depression_times = _calculate_stdp_times(
        pre_spikes, post_spikes, plastic_delay)

    # Work out the weight according to the multiplicative rule
    potentiations = (max_weight - initial_weight) * a_plus * numpy.exp(
        potentiation_times / tau_plus)
    depressions = (initial_weight - min_weight) * a_minus * numpy.exp(
        depression_times / tau_minus)
    return initial_weight + numpy.sum(potentiations) - numpy.sum(depressions)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import logging
from typing import Any, cast, Optional, Sequence, Tuple, Set, TYPE_CHECKING

import numpy

from pyNN.random import RandomDistribution

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.log import FormatAdapter

from pacman.model.graphs import AbstractVertex
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.machine import MachineVertex

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.types import (
    DELAYS, WEIGHTS_DELAYS_IN, WEIGHTS)
from spynnaker.pyNN.utilities.constants import POP_TABLE_MAX_ROW_LENGTH
from spynnaker.pyNN.exceptions import InvalidParameterType
from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    NUMPY_CONNECTORS_DTYPE as CONNECTOR_DTYPE)
from spynnaker.pyNN.types import is_scalar

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections.connectors import (
        AbstractConnector)
    from spynnaker.pyNN.models.neural_projections import SynapseInformation
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge)

logger = FormatAdapter(logging.getLogger(__name__))


class AbstractSynapseDynamics(object, metaclass=AbstractBase):
    """
    How do the dynamics of a synapse interact with the rest of the model.
    """

    __slots__ = ("__delay", "__weight")

    def __init__(self, delay: WEIGHTS_DELAYS_IN,
                 weight: WEIGHTS_DELAYS_IN):
        """
        :param delay: The delay or a way of generating the delays
        :param weight: The weights or way to generate the weights
        """
        self.__check_in_type(delay, "delay")
        self.__delay = self._round_delay(delay)
        self.__check_out_delay(self.__delay, "delay")
        self.__check_in_type(weight, "weight")
        self.__weight = self._convert_weight(weight)
        self.__check_out_weight(self.__weight, "weight")

    def __check_in_type(self, value: WEIGHTS_DELAYS_IN, name: str) -> None:
        if value is None:
            return
        if isinstance(value, (int, float, str, RandomDistribution)):
            return
        try:
            for x in value:
                if not isinstance(x, (int, numpy.integer, float)):
                    raise TypeError(
                        f"Unexpected collection of type  {type(x)} for {name}"
                        f"Expected types in collection are int and float")
            return
        except TypeError:
            # OK not a collection
            pass
        raise TypeError(
            f"Unexpected type for {name}: {type(value)}. "
            "Expected types are int, float, str, RandomDistribution "
            "and collections of type int or float")

    def __check_out_weight(self, weight: WEIGHTS, name: str) -> None:
        if weight is None:
            return
        if isinstance(weight, (int, float, str, RandomDistribution)):
            return
        if isinstance(weight, numpy.ndarray):
            for x in weight:
                if not isinstance(x, (numpy.float64)):
                    raise TypeError(
                        f"Unexpected numpy ndarray of type {type(x)}"
                        f" for {name}")
            return
        raise TypeError(
            f"Unexpected type for output data: {type(weight)} for {name} "
            "Expected types are float, str, RandomDistribution "
            "and list of type float")

    def __check_out_delay(self, delay: DELAYS, name: str) -> None:
        if isinstance(delay, (float, (str, RandomDistribution))):
            return
        if isinstance(delay, numpy.ndarray):
            for x in delay:
                if not isinstance(x, (numpy.float64)):
                    raise TypeError(
                        f"Unexpected numpy ndarray of type {type(x)}"
                        f" for {name}")
            return
        raise TypeError(
            f"Unexpected type for output data: {type(delay)} for {name} "
            "Expected types are float, str, RandomDistribution "
            "and list of type float")

    #: Type model of the basic configuration data of a connector
    NUMPY_CONNECTORS_DTYPE = CONNECTOR_DTYPE

    @abstractmethod
    def merge(self, synapse_dynamics: AbstractSynapseDynamics
              ) -> AbstractSynapseDynamics:
        """
        Merge with the given synapse_dynamics and return the result, or
        error if merge is not possible.

        :param synapse_dynamics:
        :returns: A merge of this and the given synapse_dynamics
        """
        raise NotImplementedError

    @abstractmethod
    def get_vertex_executable_suffix(self) -> str:
        """
        :returns: The executable suffix for a vertex for this dynamics.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def changes_during_run(self) -> bool:
        """
        Whether the synapses change during a run.
        """
        raise NotImplementedError

    @property
    def weight(self) -> WEIGHTS:
        """
        The weight of connections.
        """
        return self.__weight

    def _round_delay(self, delay: WEIGHTS_DELAYS_IN) -> DELAYS:
        """
        Round the delays to multiples of full timesteps.

        (otherwise SDRAM estimation calculations can go wrong)

        :param delay:
        :return: Rounded delay
        """
        if isinstance(delay, (RandomDistribution, str)):
            return delay
        if delay is None:
            delay = SpynnakerDataView.get_min_delay()
        # Note the cast is just to say trust use the delay will work
        # If not numpy will raise an exception
        new_delay = (
                numpy.rint(numpy.array(cast(float, delay)) *
                           SpynnakerDataView.get_simulation_time_step_per_ms())
                * SpynnakerDataView.get_simulation_time_step_ms())
        if not numpy.allclose(cast(float, delay), new_delay):
            logger.warning("Rounding up delay in f{} from {} to {}",
                           self, delay, new_delay)
        if isinstance(new_delay, numpy.float64):
            return float(new_delay)
        if isinstance(new_delay, numpy.ndarray):
            return new_delay
        raise TypeError(f"{type(delay)=}")

    def _convert_weight(self, weight: WEIGHTS_DELAYS_IN) -> WEIGHTS:
        """
        Convert the weights if numerical to (list of) float .

        :param weight:
        :return: weight as float (if numerical)
        """
        if weight is None:
            return weight
        if isinstance(weight, (RandomDistribution, str)):
            return weight
        if isinstance(weight, int):
            return weight
        if is_scalar(weight):
            return float(weight)
        new_weight = numpy.array(weight, dtype=float)
        return new_weight

    @property
    def delay(self) -> DELAYS:
        """
        The delay of connections.
        """
        return self.__delay

    @property
    @abstractmethod
    def is_combined_core_capable(self) -> bool:
        """
        Whether the synapse dynamics can run on a core combined with
        the neuron, or if a separate core is needed.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def is_split_core_capable(self) -> bool:
        """
        Whether the synapse dynamics can run on a core split from
        the neuron, or if only a combined core is possible.
        """
        raise NotImplementedError

    def get_synapse_parameter_names(self) -> Set[str]:
        """
        :return: the names of the parameters that can be extracted from
         synapses read from the machine.
        """
        return {"source", "target", "weight", "delay"}

    def get_value(self, key: str) -> Any:
        """
        Get a property.

        :param key: the name of the property
        :returns: Value for this key
        """
        if hasattr(self, key):
            return getattr(self, key)
        raise InvalidParameterType(
            f"Type {type(self)} does not have parameter {key}")

    def set_value(self, key: str, value: Any) -> None:
        """
        Set a property.

        :param key: the name of the parameter to change
        :param value: the new value of the parameter to assign
        """
        if hasattr(self, key):
            setattr(self, key, value)
            SpynnakerDataView.set_requires_mapping()
        else:
            raise InvalidParameterType(
                f"Type {type(self)} does not have parameter {key}")

    def get_delay_maximum(
            self, connector: AbstractConnector,
            synapse_info: SynapseInformation) -> float:
        """
        :returns: The maximum delay for the synapses.
        """
        return connector.get_delay_maximum(synapse_info)

    def get_delay_minimum(
            self, connector: AbstractConnector,
            synapse_info: SynapseInformation) -> Optional[float]:
        """
        Get the minimum delay for the synapses.

        This will support the filtering of the undelayed edge
        from the graph, but requires fixes in the synaptic manager to
        happen first before this can be utilised fully.

        :param connector: connector
        :param synapse_info: synapse info
        :returns: The minimum delay
        """
        return connector.get_delay_minimum(synapse_info)

    def get_delay_variance(
            self, connector: AbstractConnector, delays: DELAYS,
            synapse_info: SynapseInformation) -> float:
        """
        :param connector:
        :param delays:
        :param synapse_info:
        :returns: The variance in delay for the synapses.
        """
        return connector.get_delay_variance(delays, synapse_info)

    def get_weight_mean(
            self, connector: AbstractConnector,
            synapse_info: SynapseInformation) -> float:
        """
        :param connector:
        :param synapse_info:
        :returns: The mean weight for the synapses.
        """
        return connector.get_weight_mean(synapse_info.weights, synapse_info)

    def get_weight_maximum(
            self, connector: AbstractConnector,
            synapse_info: SynapseInformation) -> float:
        """
        :param connector:
        :param synapse_info:
        :returns: The maximum weight for the synapses.
        """
        return connector.get_weight_maximum(synapse_info)

    def get_weight_variance(
           self, connector: AbstractConnector, weights: WEIGHTS,
            synapse_info: SynapseInformation) -> float:
        """
        :param connector:
        :param weights:
        :param synapse_info:
        :returns: The variance in weight for the synapses.
        """
        return connector.get_weight_variance(weights, synapse_info)

    def get_synapse_id_by_target(self, target: str) -> Optional[int]:
        """
        :param target: The name of the synapse
        :returns: The index of the synapse type based on the name,
            or `None` if the name is not found.
        """
        _ = target
        return None

    def get_connected_vertices(
            self, s_info: SynapseInformation,
            source_vertex: ApplicationVertex,
            target_vertex: ApplicationVertex) -> Sequence[
                Tuple[MachineVertex, Sequence[AbstractVertex]]]:
        """
        Get the machine vertices that are connected to each other with
        this connector.

        :param s_info: The synapse information of the connection
        :param source_vertex: The source of the spikes
        :param target_vertex: The target of the spikes
        :return: A list of tuples of (target machine vertex, list of sources)
        """
        # By default, just ask the connector
        return s_info.connector.get_connected_vertices(
            s_info, source_vertex, target_vertex)

    @property
    def absolute_max_atoms_per_core(self) -> int:
        """
        The absolute maximum number of atoms per core supported by this
        synapse dynamics object.
        """
        # By default, we can only support the maximum row length per core
        return POP_TABLE_MAX_ROW_LENGTH

    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        """
        Checks that the edge supports the connector.  Returns nothing; it
        is assumed that an Exception will be raised if anything is wrong.

        By default this checks only that the views are not used
        on multi-dimensional vertices.

        :param application_edge: The edge of the connection
        :param synapse_info: The synaptic information
        """
        # By default, just ask the connector
        synapse_info.connector.validate_connection(
            application_edge, synapse_info)

    @property
    @abstractmethod
    def synapses_per_second(self) -> int:
        """
        Approximate number of synapses that can be processed per second;
        ideally as close to the truth as possible, but underestimate would
        be OK.  Overestimation would potentially mean having to handle more
        spikes than is possible, so overruns would occur.
        """
        raise NotImplementedError



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import math
from typing import Optional, Sequence, Tuple, TYPE_CHECKING

import numpy
from numpy import integer, floating, uint32
from numpy.typing import NDArray

from pyNN.random import RandomDistribution

from spinn_utilities.overrides import overrides
from spinn_utilities.safe_eval import SafeEval

from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.machine import MachineVertex
from pacman.model.graphs.common import Slice

from spinn_front_end_common.utilities.exceptions import ConfigurationException

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_machine import (
    AbstractGenerateConnectorOnMachine, ConnectorIDs)
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import SynapseInformation

_expr_context = SafeEval(
    math, numpy, numpy.arccos, numpy.arcsin, numpy.arctan, numpy.arctan2,
    numpy.ceil, numpy.cos, numpy.cosh, numpy.exp, numpy.fabs, numpy.floor,
    numpy.fmod, numpy.hypot, numpy.ldexp, numpy.log, numpy.log10, numpy.modf,
    numpy.power, numpy.sin, numpy.sinh, numpy.sqrt, numpy.tan, numpy.tanh,
    numpy.maximum, numpy.minimum, e=numpy.e, pi=numpy.pi)


class OneToOneConnector(AbstractGenerateConnectorOnMachine,
                        AbstractGenerateConnectorOnHost):
    """
    Where the pre- and postsynaptic populations have the same size,
    connect cell *i* in the presynaptic population to cell *i* in
    the postsynaptic population, for all *i*.
    """
    __slots__ = ()

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_maximum(
            synapse_info.delays,
            max(synapse_info.n_pre_neurons, synapse_info.n_post_neurons),
            synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_minimum(
            synapse_info.delays,
            max(synapse_info.n_pre_neurons, synapse_info.n_post_neurons),
            synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        delays = synapse_info.delays

        if min_delay is None or max_delay is None or delays is None:
            return 1

        if isinstance(delays, str):
            d = self._get_distances(delays, synapse_info)
            delays = _expr_context.eval(delays, d=d)
            if ((min_delay <= min(delays) <= max_delay) and (
                    min_delay <= max(delays) <= max_delay)):
                return 1
            else:
                return 0
        if isinstance(delays, (int, float, integer, floating)):
            return int(min_delay <= delays <= max_delay)
        if isinstance(delays, RandomDistribution):
            return 1

        slice_min_delay = min(delays)
        slice_max_delay = max(delays)
        return int((min_delay <= slice_max_delay <= max_delay) or
                   (min_delay <= slice_min_delay <= max_delay))

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        return 1

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_weight_maximum(
            synapse_info.weights,
            max(synapse_info.n_pre_neurons, synapse_info.n_post_neurons),
            synapse_info)

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        # Get each pre_vertex id for each post_vertex id
        post_atoms = post_vertex_slice.get_raster_ids()
        pre_atoms = numpy.array(post_atoms)

        # Filter out things where there isn't a cross over
        atom_filter = numpy.ones(len(post_atoms), dtype=numpy.bool_)
        if synapse_info.prepop_is_view or synapse_info.postpop_is_view:
            # If a view, we only keep things that are in the view
            if synapse_info.prepop_is_view:
                # pylint: disable=protected-access
                pre_lo, pre_hi = synapse_info.pre_population._view_range
                atom_filter &= (pre_atoms <= pre_hi & pre_atoms >= pre_lo)
            if synapse_info.postpop_is_view:
                # pylint: disable=protected-access
                post_lo, post_hi = synapse_info.post_population._view_range
                atom_filter &= (post_atoms <= post_hi & post_atoms >= post_lo)
        else:
            # If not a view we only keep things that are in the pre-population
            atom_filter &= (pre_atoms <= synapse_info.pre_population.size)
        post_atoms = post_atoms[atom_filter]
        pre_atoms = pre_atoms[atom_filter]

        # Convert to the correct coordinates
        post_atoms = post_vertex_slice.get_relative_indices(post_atoms)
        pre_atoms = synapse_info.pre_vertex.get_key_ordered_indices(pre_atoms)

        n_connections = len(post_atoms)
        block = numpy.zeros(n_connections, dtype=self.NUMPY_SYNAPSES_DTYPE)
        block["source"] = pre_atoms
        block["target"] = post_atoms
        block["weight"] = self._generate_weights(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self) -> str:
        return "OneToOneConnector()"

    @property
    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_id)
    def gen_connector_id(self) -> int:
        return ConnectorIDs.ONE_TO_ONE_CONNECTOR.value

    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_params)
    def gen_connector_params(
            self, synapse_info: SynapseInformation) -> NDArray[uint32]:
        return numpy.array([], dtype="uint32")

    @property
    @overrides(
        AbstractGenerateConnectorOnMachine.gen_connector_params_size_in_bytes)
    def gen_connector_params_size_in_bytes(self) -> int:
        return 0

    @overrides(AbstractGenerateConnectorOnMachine.get_connected_vertices)
    def get_connected_vertices(
            self, s_info: SynapseInformation, source_vertex: ApplicationVertex,
            target_vertex: ApplicationVertex) -> Sequence[
                Tuple[MachineVertex, Sequence[MachineVertex]]]:
        src_vtxs = source_vertex.splitter.get_out_going_vertices(
            s_info.partition_id)
        tgt_vtxs = target_vertex.splitter.get_in_coming_vertices(
            s_info.partition_id)

        # If doing a view, we must be single dimensional, so use old method
        if (len(s_info.pre_vertex.atoms_shape) == 1 and
                len(s_info.post_vertex.atoms_shape) == 1):

            pre_lo = 0
            pre_hi = source_vertex.n_atoms - 1
            post_lo = 0
            post_hi = target_vertex.n_atoms - 1
            if s_info.prepop_is_view:
                # pylint: disable=protected-access
                pre_lo, pre_hi = s_info.pre_population._view_range
            if s_info.postpop_is_view:
                # pylint: disable=protected-access
                post_lo, post_hi = s_info.post_population._view_range

            return [(t_vert,
                     [s_vert for s_vert in src_vtxs if self.__connects(
                          s_vert, pre_lo, pre_hi, t_vert, post_lo, post_hi)])
                    for t_vert in tgt_vtxs]

        if s_info.prepop_is_view or s_info.postpop_is_view:

            # Check again here in case the rules change elsewhere
            raise ConfigurationException(
                "The OneToOneConnector does not support PopulationView "
                "connections between vertices with more than 1 dimension")

        # Check for cross over of pre- and post- rasters, as that is how the
        # connector works
        return [(t_vert,
                 [s_vert for s_vert in src_vtxs if any(numpy.isin(
                     s_vert.vertex_slice.get_raster_ids(),
                     t_vert.vertex_slice.get_raster_ids()))])
                for t_vert in tgt_vtxs]

    def __connects(
            self, s_vert: MachineVertex, pre_lo: int, pre_hi: int,
            t_vert: MachineVertex, post_lo: int, post_hi: int) -> bool:
        pre_slice = s_vert.vertex_slice
        post_slice = t_vert.vertex_slice

        # Check range of slices
        if pre_slice.hi_atom < pre_lo:
            return False
        if post_slice.hi_atom < post_lo:
            return False
        if pre_slice.lo_atom > pre_hi:
            return False
        if post_slice.lo_atom > post_hi:
            return False

        # Get slice range relative to view
        pre_s_hi = pre_slice.hi_atom - pre_lo
        post_s_hi = post_slice.hi_atom - post_lo
        pre_s_lo = pre_slice.lo_atom - pre_lo
        post_s_lo = post_slice.lo_atom - post_lo

        if pre_s_hi < post_s_lo:
            return False
        if pre_s_lo > post_s_hi:
            return False

        return True

    @overrides(AbstractGenerateConnectorOnMachine.generate_on_machine)
    def generate_on_machine(self, synapse_info: SynapseInformation) -> bool:
        # If we are doing a 1:1 connector and the pre or post vertex is
        # multi-dimensional and have different dimensions
        pre = synapse_info.pre_vertex
        post = synapse_info.post_vertex
        if len(pre.atoms_shape) > 1 or len(post.atoms_shape) > 1:
            if (pre.atoms_shape != post.atoms_shape):
                print("Not generating on core!")
                return False
            if (pre.get_max_atoms_per_dimension_per_core() !=
                    post.get_max_atoms_per_dimension_per_core()):
                print("Not generating on core!")
                return False
        return super(OneToOneConnector, self).generate_on_machine(synapse_info)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from numpy import floating
from numpy.typing import NDArray
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_front_end_common.interface.ds import DataSpecificationBase
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractHasParameterNames)


class AbstractWeightDependence(
        AbstractHasParameterNames, metaclass=AbstractBase):
    """
    API with the weight dependency methods.
    """
    __slots__ = ()

    @abstractmethod
    def is_same_as(
            self, weight_dependence: "AbstractWeightDependence") -> bool:
        """
        Determine if this weight dependence is the same as another.

        :param weight_dependence:
        :returns: True if the weight dependence is the same as another.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def vertex_executable_suffix(self) -> str:
        """
        The suffix to be appended to the vertex executable for this rule.
        """
        raise NotImplementedError

    @abstractmethod
    def get_parameters_sdram_usage_in_bytes(
            self, n_synapse_types: int, n_weight_terms: int) -> int:
        """
        :param n_synapse_types:
        :param n_weight_terms:
        :returns: the amount of SDRAM used by the parameters of this rule.
        """
        raise NotImplementedError

    @abstractmethod
    def write_parameters(
            self, spec: DataSpecificationBase, global_weight_scale: float,
            synapse_weight_scales: NDArray[floating],
            n_weight_terms: int) -> None:
        """
        Write the parameters of the rule to the spec.

        :param spec: The specification to write to
        :param global_weight_scale: The weight scale applied globally
        :param synapse_weight_scales:
            The total weight scale applied to each synapse including the global
            weight scale
        :param n_weight_terms: The number of terms used by the synapse rule
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def weight_maximum(self) -> float:
        """
        The maximum weight that will ever be set in a synapse as a result
        of this rule.
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import logging
from typing import (
    Any, Callable, Dict, final, Optional, Sequence, Tuple, TYPE_CHECKING,
    Union)

from pyNN.space import BaseStructure
import neo
from numpy import floating
from numpy.typing import NDArray
from typing_extensions import Never

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.log import FormatAdapter
from spinn_utilities.logger_utils import warn_once

from pacman.model.graphs.application import ApplicationVertex

from spynnaker.pyNN.models.recorder import Recorder
from spynnaker.pyNN.types import IoDest

if TYPE_CHECKING:
    from pyNN.neuron.standardmodels.electrodes import NeuronCurrentSource
    from spynnaker.pyNN.models.common.types import Names
    from .population_view import IDMixin

logger = FormatAdapter(logging.getLogger(__name__))


class PopulationBase(object, metaclass=AbstractBase):
    r"""
    Shared methods between :py:class:`Population`\ s and
    :py:class:`PopulationView`\ s.

    Mainly pass through and not implemented.
    """
    __slots__ = ()

    @property
    def local_cells(self) -> Sequence[IDMixin]:
        """
        An array containing the cell IDs of those neurons in the
        Population that exist on the local MPI node.
        """
        logger.warning("local calls do not really make sense on sPyNNaker so "
                       "local_cells just returns all_cells")
        return self.all_cells

    @property
    @abstractmethod
    def all_cells(self) -> Sequence[IDMixin]:
        """
        An array containing the cell IDs of all neurons in the
        Population (all MPI nodes).
        """
        raise NotImplementedError

    @final
    def __add__(self, other: PopulationBase) -> Never:  # pylint: disable=W0613
        """
        A Population / PopulationView can be added to another
        Population, PopulationView or Assembly, returning an Assembly.

        .. warning::
            Currently unimplemented.

        :param other:
        """
        raise NotImplementedError("sPyNNaker does not support Assemblies")

    @abstractmethod
    def get_data(
            self, variables: Names = 'all',
            gather: bool = True, clear: bool = False, *,
            annotations: Optional[Dict[str, Any]] = None) -> neo.Block:
        """
        Return a Neo Block containing the data(spikes, state variables)
        recorded from the Population.

        :param variables:
            Either a single variable name or a list of variable names.
            Variables must have been previously recorded, otherwise an
            Exception will be raised.
        :param gather: Ignored. Purely for PyNN compatibility

            .. note::
                SpiNNaker always gathers.

        :param clear:
            If this is True, recorded data will be deleted from the Population.
        :param annotations: annotations to put on the neo block
        :returns: Data in neo format
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the variable or variables have not been previously set to
            record.
        """
        raise NotImplementedError

    @abstractmethod
    def get_spike_counts(self, gather: bool = True) -> Dict[int, int]:
        """
        Returns a dict containing the number of spikes for each neuron.

        The dict keys are neuron IDs, not indices.

        :param gather: Ignored. Purely for PyNN compatibility

            .. note::
                SpiNNaker always gathers.
        :returns: A dict containing the number of spikes for each neuron.
        """
        raise NotImplementedError

    def inject(self, current_source: NeuronCurrentSource) -> None:
        """
        Connect a current source to all cells in the Population.

        :param current_source:
        """
        raise NotImplementedError

    @final
    def is_local(self, id: int) -> bool:  # @ReservedAssignment
        """
        Indicates whether the cell with the given ID exists on the
        local MPI node.

        :returns: True
        """
        # pylint: disable=redefined-builtin
        _ = id
        logger.warning("local calls do not really make sense on sPyNNaker so "
                       "is_local always returns True")
        return True

    @property
    @final
    def local_size(self) -> int:
        """
        The number of cells in the population on the local MPI node.
        """
        logger.warning("local calls do not really make sense on sPyNNaker so "
                       "is_local always returns size")
        return len(self)

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    def mean_spike_count(self, gather: bool = True) -> float:
        """
        :param gather: Ignored. Purely for PyNN compatibility

            .. note::
                SpiNNaker always gathers.
        :returns: The mean number of spikes per neuron.
        """
        if not gather:
            warn_once(
                logger, "sPyNNaker only supports gather=True. We will run "
                "as if gather was set to True.")
        counts = self.get_spike_counts()
        return sum(counts.values()) / len(counts)

    @final
    def nearest(
            self, position: Any) -> Never:
        """
        Return the neuron closest to the specified position.

        .. warning::
            Currently unimplemented.
        """
        _ = position
        raise NotImplementedError("sPyNNaker does not currently do nearest")

    @property
    @abstractmethod
    def position_generator(self) -> Callable[[int], NDArray[floating]]:
        """
        Returns a function to get the position for single id
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def positions(self) -> NDArray[floating]:
        """
        The position array for structured populations.

        This will be a 3xN array
        """
        raise NotImplementedError

    @abstractmethod
    def write_data(self, io: Union[str, neo.baseio.BaseIO],
                   variables: Names = 'all',
                   gather: bool = True, clear: bool = False,
                   annotations: Optional[Dict[str, Any]] = None) -> None:
        """
        Write recorded data to file, using one of the file formats
        supported by Neo.

        :param io:
            a Neo IO instance, or a string for where to put a Neo instance
        :param variables:
            either a single variable name or a list of variable names.
            Variables must have been previously recorded, otherwise an
            Exception will be raised.
        :param gather: Ignored. Purely for PyNN compatibility

            .. note::
                SpiNNaker always gathers.

        :param clear:
            clears the storage data if set to true after reading it back
        :param annotations: annotations to put on the Neo block
        """
        raise NotImplementedError

    @final
    def receptor_types(self) -> Never:
        """
        .. note::
            NO PyNN description of this method.

        .. warning::
            Currently unimplemented.
        """
        raise NotImplementedError()

    @abstractmethod
    def record(self, variables: Names, to_file: IoDest = None,
               sampling_interval: Optional[int] = None) -> None:
        """
        Record the specified variable or variables for all cells in the
        Population or view.

        :param variables: either a single variable name or a list of variable
            names. For a given `celltype` class, `celltype.recordable` contains
            a list of variables that can be recorded for that `celltype`.
        :param to_file: a file to automatically record to (optional).
            `write_data()` will be automatically called when `end()` is called.
        :param sampling_interval: a value in milliseconds, and an integer
            multiple of the simulation timestep.
        """
        raise NotImplementedError

    @final
    def save_positions(self, file: Any) -> Never:  # pylint: disable=W0622
        """
        Save positions to file. The output format is index x y z

        .. warning::
            Currently unimplemented.
        """
        _ = file
        raise NotImplementedError

    @property
    @abstractmethod
    def structure(self) -> Optional[BaseStructure]:
        """
        The spatial structure of the parent Population.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _vertex(self) -> ApplicationVertex:
        """
        The underlying application vertex.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _view_range(self) -> Tuple[int, int]:
        """
        The range of neuron IDs supported.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _recorder(self) -> Recorder:
        """
        The recorder of the population.
        """
        raise NotImplementedError

    @staticmethod
    def _check_params(gather: bool,
                      annotations: Optional[Dict[str, Any]] = None) -> None:
        if not gather:
            logger.warning(
                "sPyNNaker only supports gather=True. We will run "
                "as if gather was set to True.")
        if annotations is not None:
            warn_once(
                logger, "annotations parameter is not standard PyNN so may "
                        "not be supported by all platforms.")



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spynnaker.pyNN.models.projection import Projection


class AbstractSupportsOneToOneSDRAMInput(metaclass=AbstractBase):
    """
    An interface for a splitter that supports one-to-one input using
    SDRAM.  The splitter is assumed to handle the splitting on any inputs
    that are actually one-to-one, as it will have to create the vertices.
    """

    @abstractmethod
    def handles_source_vertex(self, projection: Projection) -> bool:
        """
        Determine if the source vertex of the given projection is to be
        handled by the target splitter.

        :param  projection: The projection to check the source of
        :returns: True if the target can handle this projection
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, List, Optional

from spinn_utilities.overrides import overrides

from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utility_models import MultiCastCommand

from spynnaker.pyNN.external_devices_models.push_bot.parameters import (
    PushBotMotor)
from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol
from .push_bot_device import PushBotEthernetDevice


class PushBotEthernetMotorDevice(
        PushBotEthernetDevice, AbstractSendMeMulticastCommandsVertex):
    """
    The motor of a PushBot.
    """

    def __init__(self, motor: PushBotMotor,
                 protocol: MunichIoSpiNNakerLinkProtocol,
                 timesteps_between_send: Optional[int] = None):
        """
        :param motor: indicates which motor to control
        :param protocol:
            The protocol used to control the device
        :param timesteps_between_send:
            The number of timesteps between sending commands to the device,
            or `None` to use the default
        """
        if not isinstance(motor, PushBotMotor):
            raise ConfigurationException(
                "motor parameter must be a PushBotMotor value")

        super().__init__(protocol, motor, True, timesteps_between_send)
        self.__command_protocol = protocol

    @overrides(PushBotEthernetDevice.set_command_protocol)
    def set_command_protocol(
            self, command_protocol: MunichIoSpiNNakerLinkProtocol) -> None:
        self.__command_protocol = command_protocol

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.start_resume_commands)
    def start_resume_commands(self) -> Iterable[MultiCastCommand]:
        # add mode command if not done already
        if not self.protocol.sent_mode_command():
            yield self.protocol.set_mode()

        # device specific commands
        yield self.__command_protocol.generic_motor_enable()

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.pause_stop_commands)
    def pause_stop_commands(self) -> Iterable[MultiCastCommand]:
        yield self.__command_protocol.generic_motor_disable()

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.timed_commands)
    def timed_commands(self) -> List[MultiCastCommand]:
        return []



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple

from spinn_utilities.overrides import overrides
from spinn_utilities.ranged import RangeDictionary

from spinn_front_end_common.interface.ds import DataType

from spynnaker.pyNN.models.neuron.implementations import ModelParameter
from spynnaker.pyNN.utilities.struct import Struct
from spynnaker.pyNN.data import SpynnakerDataView

from .abstract_synapse_type import AbstractSynapseType

EXC_RESPONSE = "exc_response"
EXC_EXP_RESPONSE = "exc_exp_response"
TAU_SYN_E = "tau_syn_E"
INH_RESPONSE = "inh_response"
INH_EXP_RESPONSE = "inh_exp_response"
TAU_SYN_I = "tau_syn_I"
Q_EXC = "q_exc"
Q_INH = "q_inh"
TIMESTEP_MS = "timestep_ms"


class SynapseTypeAlpha(AbstractSynapseType):
    """
    Alpha Type Synapse
    """
    __slots__ = (
        "__exc_exp_response",
        "__exc_response",
        "__inh_exp_response",
        "__inh_response",
        "__tau_syn_E",
        "__tau_syn_I",
        "__q_exc",
        "__q_inh")

    def __init__(
            self, exc_response: ModelParameter,
            exc_exp_response: ModelParameter, tau_syn_E: ModelParameter,
            inh_response: ModelParameter, inh_exp_response: ModelParameter,
            tau_syn_I: ModelParameter):
        r"""
        :param exc_response: :math:`response^\mathrm{linear}_e`
        :param exc_exp_response: :math:`response^\mathrm{exponential}_e`
        :param tau_syn_E: :math:`\tau^{syn}_e`
        :param inh_response: :math:`response^\mathrm{linear}_i`
        :param inh_exp_response: :math:`response^\mathrm{exponential}_i`
        :param tau_syn_I: :math:`\tau^{syn}_i`
        """
        super().__init__(
            [Struct([
                (DataType.S1615, EXC_RESPONSE),
                (DataType.S1615, EXC_EXP_RESPONSE),
                (DataType.S1615, Q_EXC),
                (DataType.S1615, TAU_SYN_E),
                (DataType.S1615, INH_RESPONSE),
                (DataType.S1615, INH_EXP_RESPONSE),
                (DataType.S1615, Q_INH),
                (DataType.S1615, TAU_SYN_I),
                (DataType.S1615, TIMESTEP_MS)])],
            {EXC_RESPONSE: "", EXC_EXP_RESPONSE: "", TAU_SYN_E: "ms",
             INH_RESPONSE: "", INH_EXP_RESPONSE: "", TAU_SYN_I: "ms"})

        self.__exc_response = exc_response
        self.__exc_exp_response = exc_exp_response
        # pylint: disable=invalid-name
        self.__tau_syn_E = tau_syn_E
        self.__inh_response = inh_response
        self.__inh_exp_response = inh_exp_response
        self.__tau_syn_I = tau_syn_I

    @overrides(AbstractSynapseType.add_parameters)
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        parameters[TAU_SYN_E] = self._convert(self.__tau_syn_E)
        parameters[TAU_SYN_I] = self._convert(self.__tau_syn_I)

    @overrides(AbstractSynapseType.add_state_variables)
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        state_variables[EXC_RESPONSE] = self._convert(self.__exc_response)
        state_variables[EXC_EXP_RESPONSE] = self._convert(
            self.__exc_exp_response)
        state_variables[Q_EXC] = 0
        state_variables[INH_RESPONSE] = self._convert(self.__inh_response)
        state_variables[INH_EXP_RESPONSE] = self._convert(
            self.__inh_exp_response)
        state_variables[Q_INH] = 0
        state_variables[TIMESTEP_MS] = (
            SpynnakerDataView.get_simulation_time_step_ms())

    @overrides(AbstractSynapseType.get_n_synapse_types)
    def get_n_synapse_types(self) -> int:
        return 2  # excitatory and inhibitory

    @overrides(AbstractSynapseType.get_synapse_id_by_target)
    def get_synapse_id_by_target(self, target: str) -> Optional[int]:
        if target == "excitatory":
            return 0
        elif target == "inhibitory":
            return 1
        return None

    @overrides(AbstractSynapseType.get_synapse_targets)
    def get_synapse_targets(self) -> Tuple[str, ...]:
        return "excitatory", "inhibitory"

    @property
    def exc_response(self) -> ModelParameter:
        """
        Value as passed into the init.
        """
        return self.__exc_response

    @exc_response.setter
    def exc_response(self, exc_response: ModelParameter) -> None:
        """
        Value as passed into the init.
        """
        self.__exc_response = exc_response

    @property
    def tau_syn_E(self) -> ModelParameter:
        """
        Value as passed into the init.
        """
        # pylint: disable=invalid-name
        return self.__tau_syn_E

    @property
    def inh_response(self) -> ModelParameter:
        """
        Value as passed into the init.
        """
        return self.__inh_response

    @property
    def tau_syn_I(self) -> ModelParameter:
        """
        Value as passed into the init.
        """
        # pylint: disable=invalid-name
        return self.__tau_syn_I



# Copyright (c) 2024 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Iterable, List, Optional, Tuple, Dict, TYPE_CHECKING, cast
import logging

import numpy
from numpy import floating, integer, uint8, uint16, uint32
from numpy.typing import NDArray

from pyNN.standardmodels.synapses import StaticSynapse

from spinn_utilities.overrides import overrides
from spinn_utilities.log import FormatAdapter

from spinn_front_end_common.interface.ds import DataSpecificationBase
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_WORD, BYTES_PER_SHORT)

from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.neural_projections.connectors import (
    AbstractConnector)
from spynnaker.pyNN.types import WEIGHTS
from spynnaker.pyNN.types import WEIGHTS_DELAYS_IN as _In_Types
from spynnaker.pyNN.utilities.utility_calls import get_n_bits
from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    NUMPY_CONNECTORS_DTYPE)
from spynnaker.pyNN.models.neural_projections.connectors import (
    AbstractGenerateConnectorOnMachine)
from .abstract_plastic_synapse_dynamics import AbstractPlasticSynapseDynamics
from .abstract_generate_on_machine import (
    AbstractGenerateOnMachine, MatrixGeneratorID)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
        ConnectionsArray)
    from spynnaker.pyNN.models.neuron.synapse_io import MaxRowInfo
    from .abstract_synapse_dynamics import AbstractSynapseDynamics

logger = FormatAdapter(logging.getLogger(__name__))


class SynapseDynamicsWeightChangable(
        AbstractPlasticSynapseDynamics, AbstractGenerateOnMachine):
    """
    The dynamics of a synapse that can be changed simply by the sending of an
    external signal.
    """

    __slots__ = (

        # The maximum weight
        "__weight_max",

        # The minimum weight
        "__weight_min",

        # The map of synapse information to index
        "__synapse_info_to_index",

        # The next index to use for the next projection
        "__next_index")

    def __init__(
            self,
            weight_min: float, weight_max: float,
            weight: _In_Types = StaticSynapse.default_parameters['weight'],
            delay: _In_Types = None):
        """
        :param weight:
        :param delay: Use ``None`` to get the simulator default minimum delay.
        """
        super().__init__(delay=delay, weight=weight)
        self.__weight_max = weight_max
        self.__weight_min = weight_min
        self.__synapse_info_to_index: Dict[SynapseInformation, int] = dict()
        self.__next_index = 0

        if weight_min < 0.0:
            raise SynapticConfigurationException(
                "The minimum weight must be greater than or equal to 0")
        if weight_max < 0.0:
            raise SynapticConfigurationException(
                "The maximum weight must be greater than or equal to 0")
        if weight_min >= weight_max:
            raise SynapticConfigurationException(
                "The minimum weight must be less than the maximum")

    @property
    def weight_max(self) -> float:
        """ Get the maximum weight allowed to change to
        """
        return self.__weight_max

    @property
    def weight_min(self) -> float:
        """ Get the minimum weight allowed to change to
        """
        return self.__weight_min

    def get_synapse_info_index(self, synapse_info: SynapseInformation) -> int:
        """
        Get the row offset for the given synapse information.

        Each synapse information has a unique row offset which then allows
        for multiple connections to be identified and kept separate.

        :returns: The row offset for the given synapse information.
        """
        if synapse_info not in self.__synapse_info_to_index:
            self.__synapse_info_to_index[synapse_info] = self.__next_index
            self.__next_index += synapse_info.pre_vertex.n_atoms
        return self.__synapse_info_to_index[synapse_info]

    @overrides(AbstractPlasticSynapseDynamics.merge)
    def merge(self, synapse_dynamics: AbstractSynapseDynamics
              ) -> AbstractSynapseDynamics:
        # If dynamics is a WeightChanger, return ourselves, as
        # WeightChanger can't be used by itself
        # Note: hack required to avoid circular import
        # pylint: disable=import-outside-toplevel
        from .synapse_dynamics_weight_changer import (
            SynapseDynamicsWeightChanger)
        if isinstance(synapse_dynamics, SynapseDynamicsWeightChanger):
            return self

        if not isinstance(synapse_dynamics, SynapseDynamicsWeightChangable):
            raise SynapticConfigurationException(
                "Only a WeightChanger and WeightChangable can be combined")

        if not self.is_same_as(synapse_dynamics):
            raise SynapticConfigurationException(
                "Multiple WeightChangables must have the same min and max")
        return self

    @overrides(AbstractPlasticSynapseDynamics.is_same_as)
    def is_same_as(self, synapse_dynamics: AbstractSynapseDynamics) -> bool:
        if not isinstance(synapse_dynamics, SynapseDynamicsWeightChangable):
            return False
        return (synapse_dynamics.weight_max == self.weight_max and
                synapse_dynamics.weight_min == self.weight_min)

    @overrides(AbstractPlasticSynapseDynamics.get_vertex_executable_suffix)
    def get_vertex_executable_suffix(self) -> str:
        return "_weight_change"

    @overrides(AbstractPlasticSynapseDynamics.
               get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(
            self, n_neurons: int, n_synapse_types: int) -> int:
        # The count of items, plus min and max for each synapse type
        return BYTES_PER_WORD + (BYTES_PER_SHORT * 2 * n_synapse_types)

    @overrides(AbstractPlasticSynapseDynamics.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationBase, region: int,
            global_weight_scale: float,
            synapse_weight_scales: NDArray[floating]) -> None:
        spec.comment("Writing Plastic Parameters")

        # Switch focus to the region:
        spec.switch_write_focus(region)

        spec.write_value(len(synapse_weight_scales))
        min_weights = (
            synapse_weight_scales * global_weight_scale *
            self.__weight_min)
        max_weights = (
            synapse_weight_scales * global_weight_scale *
            self.__weight_max)
        weights = numpy.dstack((min_weights, max_weights)).flatten().astype(
            uint16).view(uint32)
        spec.write_array(weights)

    @overrides(AbstractPlasticSynapseDynamics.
               get_n_words_for_plastic_connections)
    def get_n_words_for_plastic_connections(self, n_connections: int) -> int:
        """
        :param n_connections:
        """
        n_words = n_connections // 2
        if n_connections % 2 != 0:
            n_words += 1
        # plastic-plastic has 1 header and then a half-word-weight per
        # connection
        # fixed-plastic has a half-word per connection for the other elements
        return 1 + n_words * 2

    @overrides(AbstractPlasticSynapseDynamics.get_plastic_synaptic_data)
    def get_plastic_synaptic_data(
            self, connections: ConnectionsArray,
            connection_row_indices: NDArray[integer], n_rows: int,
            n_synapse_types: int,
            max_n_synapses: int, max_atoms_per_core: int) -> Tuple[
                List[NDArray[uint32]], List[NDArray[uint32]],
                NDArray[uint32], NDArray[uint32]]:
        raise NotImplementedError(
            "WeightChangable can only be generated on machine")

    @overrides(
        AbstractPlasticSynapseDynamics.get_n_plastic_plastic_words_per_row)
    def get_n_plastic_plastic_words_per_row(
            self, pp_size: NDArray[uint32]) -> NDArray[integer]:
        # pp_size is in words, so return
        return pp_size

    @overrides(
        AbstractPlasticSynapseDynamics.get_n_fixed_plastic_words_per_row)
    def get_n_fixed_plastic_words_per_row(
            self, fp_size: NDArray[uint32]) -> NDArray[integer]:
        # fp_size is in half-words
        return numpy.ceil(fp_size / 2.0).astype(dtype=uint32)

    @overrides(AbstractPlasticSynapseDynamics.get_n_synapses_in_rows)
    def get_n_synapses_in_rows(self, pp_size: NDArray[uint32],
                               fp_size: NDArray[uint32]) -> NDArray[integer]:
        # Each fixed-plastic synapse is a half-word and fp_size is in half
        # words so just return it
        return fp_size

    @overrides(AbstractPlasticSynapseDynamics.read_plastic_synaptic_data)
    def read_plastic_synaptic_data(
            self, n_synapse_types: int, pp_size: NDArray[uint32],
            pp_data: List[NDArray[uint32]], fp_size: NDArray[uint32],
            fp_data: List[NDArray[uint32]],
            max_atoms_per_core: int) -> ConnectionsArray:
        logger.warning(
            "Weights are only changed when a pre-spike arrives after a"
            " change-spike has been received, and so the weights might"
            " not be as expected")

        n_rows = len(fp_size)

        n_synapse_type_bits = get_n_bits(n_synapse_types)
        n_neuron_id_bits = get_n_bits(max_atoms_per_core)
        neuron_id_mask = (1 << n_neuron_id_bits) - 1

        data_fixed = numpy.concatenate([
            fp_data[i].view(dtype=uint16)[0:fp_size[i]]
            for i in range(n_rows)])
        pp_without_headers = [
            row.view(dtype=uint8)[BYTES_PER_WORD:] for row in pp_data]
        pp_half_words = numpy.concatenate(
            [pp[:size * BYTES_PER_SHORT]
             for pp, size in zip(pp_without_headers, fp_size)]).view(uint16)

        connections = numpy.zeros(
            data_fixed.size, dtype=NUMPY_CONNECTORS_DTYPE)
        connections["source"] = numpy.concatenate(
            [numpy.repeat(i, fp_size[i]) for i in range(len(fp_size))])
        connections["target"] = data_fixed & neuron_id_mask
        connections["weight"] = pp_half_words
        connections["delay"] = data_fixed >> (
            n_neuron_id_bits + n_synapse_type_bits)
        return connections

    @overrides(AbstractPlasticSynapseDynamics.get_weight_mean)
    def get_weight_mean(self, connector: AbstractConnector,
                        synapse_info: SynapseInformation) -> float:
        # Because the weights could all be changed to the maximum, the mean
        # has to be given as the maximum for scaling
        return self.get_weight_maximum(connector, synapse_info)

    @overrides(AbstractPlasticSynapseDynamics.get_weight_variance)
    def get_weight_variance(
           self, connector: AbstractConnector, weights: WEIGHTS,
            synapse_info: SynapseInformation) -> float:
        # Because the weights could all be changed to the maximum, the variance
        # has to be given as no variance
        return 0.0

    @overrides(AbstractPlasticSynapseDynamics.get_weight_maximum)
    def get_weight_maximum(self, connector: AbstractConnector,
                           synapse_info: SynapseInformation) -> float:
        return self.__weight_max

    @overrides(AbstractPlasticSynapseDynamics.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        yield 'weight'
        yield 'delay'

    @overrides(AbstractPlasticSynapseDynamics.get_max_synapses)
    def get_max_synapses(self, n_words: int) -> int:
        # Subtract the header size that will always exist
        n_words_space = n_words - BYTES_PER_WORD

        # The remaining space is divided equally into plastic plastic and fixed
        # plastic, but these have to be word aligned, so e.g. 5 words space
        # would be 2.5 words per section, which allows for 5 connections in
        # theory, but 5 connections would require word alignment at 3 words per
        # section, so we round down instead and get 4 connections
        return 2 * (n_words_space // 2)

    @property
    @overrides(AbstractGenerateOnMachine.gen_matrix_id)
    def gen_matrix_id(self) -> int:
        # We can use the STDP generation routine, with the addition that
        # each row header has the pre-neuron-id in it (i.e. the row number)
        return MatrixGeneratorID.STDP_MATRIX.value

    @overrides(AbstractGenerateOnMachine.gen_matrix_params)
    def gen_matrix_params(
            self, synaptic_matrix_offset: int, delayed_matrix_offset: int,
            app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation, max_row_info: MaxRowInfo,
            max_pre_atoms_per_core: int, max_post_atoms_per_core: int
            ) -> NDArray[uint32]:
        vertex = app_edge.post_vertex
        n_synapse_type_bits = get_n_bits(
            vertex.neuron_impl.get_n_synapse_types())
        n_synapse_index_bits = get_n_bits(max_post_atoms_per_core)
        max_delay = app_edge.post_vertex.splitter.max_support_delay()
        max_delay_bits = get_n_bits(max_delay)
        half_word = 0
        n_half_words = 1
        header_half_words = 2
        write_row_number_to_header = 1
        # We need to use the "global" dynamics object to get the offset
        dynamics = cast(SynapseDynamicsWeightChangable,
                        app_edge.post_vertex.synapse_dynamics)
        row_offset = dynamics.get_synapse_info_index(synapse_info)
        return numpy.array([
            synaptic_matrix_offset, delayed_matrix_offset,
            max_row_info.undelayed_max_n_synapses,
            max_row_info.delayed_max_n_synapses,
            max_row_info.undelayed_max_words, max_row_info.delayed_max_words,
            synapse_info.synapse_type, n_synapse_type_bits,
            n_synapse_index_bits, app_edge.n_delay_stages + 1,
            max_delay, max_delay_bits, app_edge.pre_vertex.n_atoms,
            max_pre_atoms_per_core, header_half_words,
            n_half_words, half_word, write_row_number_to_header, row_offset],
            dtype=uint32)

    @property
    @overrides(AbstractGenerateOnMachine.gen_matrix_params_size_in_bytes)
    def gen_matrix_params_size_in_bytes(self) -> int:
        return 19 * BYTES_PER_WORD

    @property
    @overrides(AbstractPlasticSynapseDynamics.changes_during_run)
    def changes_during_run(self) -> bool:
        return True

    @property
    @overrides(AbstractPlasticSynapseDynamics.is_combined_core_capable)
    def is_combined_core_capable(self) -> bool:
        return True

    @property
    @overrides(AbstractPlasticSynapseDynamics.is_split_core_capable)
    def is_split_core_capable(self) -> bool:
        return True

    @property
    @overrides(AbstractPlasticSynapseDynamics.pad_to_length)
    def pad_to_length(self) -> Optional[int]:
        return None

    @overrides(AbstractPlasticSynapseDynamics.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        AbstractPlasticSynapseDynamics.validate_connection(
            self, application_edge, synapse_info)
        if not isinstance(synapse_info.connector,
                          AbstractGenerateConnectorOnMachine):
            raise SynapticConfigurationException(
                "WeightChangable only works with on-machine generated "
                "connectors at present")
        if not synapse_info.connector.generate_on_machine(synapse_info):
            raise SynapticConfigurationException(
                "WeightChangable only works with on-machine generated "
                "connectors at present")

    @property
    @overrides(AbstractPlasticSynapseDynamics.synapses_per_second)
    def synapses_per_second(self) -> int:
        # From Synapse-Centric Mapping of Cortical Models to the SpiNNaker
        # Neuromorphic Architecture; but adapted since this does very little
        # in terms of execution compared to STDP, but a bit more than static
        return 10000000



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from spinn_utilities.overrides import overrides
from spinn_utilities.ranged import RangeDictionary
from spinn_front_end_common.interface.ds import DataType
from spynnaker.pyNN.models.neuron.implementations import (
    AbstractStandardNeuronComponent, ModelParameter)
from spynnaker.pyNN.utilities.struct import Struct
from spynnaker.pyNN.data import SpynnakerDataView
from .neuron_model import NeuronModel

V = "v"
V_REST = "v_rest"
TAU_M = "tau_m"
CM = "cm"
I_OFFSET = "i_offset"
V_RESET = "v_reset"
TAU_REFRAC = "tau_refrac"
TIMESTEP = "timestep"
REFRACT_TIMER = "refract_timer"


class NeuronModelLeakyIntegrateAndFire(NeuronModel):
    """
    Classic leaky integrate and fire neuron model.
    """
    __slots__ = (
        "__v_init",
        "__v_rest",
        "__tau_m",
        "__cm",
        "__i_offset",
        "__v_reset",
        "__tau_refrac")

    def __init__(
            self, v_init: Optional[ModelParameter], v_rest: ModelParameter,
            tau_m: ModelParameter, cm: ModelParameter,
            i_offset: ModelParameter, v_reset: ModelParameter,
            tau_refrac: ModelParameter):
        r"""
        :param v_init: :math:`V_{init}`
        :param v_rest: :math:`V_{rest}`
        :param tau_m: :math:`\tau_{m}`
        :param cm: :math:`C_m`
        :param i_offset: :math:`I_{offset}`
        :param v_reset: :math:`V_{reset}`
        :param tau_refrac: :math:`\tau_{refrac}`
        """
        super().__init__(
            [Struct([
                (DataType.S1615, V),
                (DataType.S1615, V_REST),
                (DataType.S1615, CM),
                (DataType.S1615, TAU_M),
                (DataType.S1615, I_OFFSET),
                (DataType.S1615, V_RESET),
                (DataType.S1615, TAU_REFRAC),
                (DataType.INT32, REFRACT_TIMER),
                (DataType.S1615, TIMESTEP)])],
            {V: 'mV', V_REST: 'mV', TAU_M: 'ms', CM: 'nF', I_OFFSET: 'nA',
             V_RESET: 'mV', TAU_REFRAC: 'ms'})

        if v_init is None:
            v_init = v_rest
        self.__v_init = v_init
        self.__v_rest = v_rest
        self.__tau_m = tau_m
        self.__cm = cm
        self.__i_offset = i_offset
        self.__v_reset = v_reset
        self.__tau_refrac = tau_refrac

    @overrides(AbstractStandardNeuronComponent.add_parameters)
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        parameters[V_REST] = self._convert(self.__v_rest)
        parameters[TAU_M] = self._convert(self.__tau_m)
        parameters[CM] = self._convert(self.__cm)
        parameters[I_OFFSET] = self._convert(self.__i_offset)
        parameters[V_RESET] = self._convert(self.__v_reset)
        parameters[TAU_REFRAC] = self._convert(self.__tau_refrac)
        parameters[TIMESTEP] = SpynnakerDataView.get_simulation_time_step_ms()

    @overrides(AbstractStandardNeuronComponent.add_state_variables)
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        state_variables[V] = self._convert(self.__v_init)
        state_variables[REFRACT_TIMER] = 0

    @property
    def v_init(self) -> ModelParameter:
        """
        Settable model parameter: :math:`V_{init}`
        """
        return self.__v_init

    @property
    def v_rest(self) -> ModelParameter:
        """
        Settable model parameter: :math:`V_{rest}`
        """
        return self.__v_rest

    @property
    def tau_m(self) -> ModelParameter:
        r"""
        Settable model parameter: :math:`\tau_{m}`
        """
        return self.__tau_m

    @property
    def cm(self) -> ModelParameter:
        """
        Settable model parameter: :math:`C_m`
        """
        return self.__cm

    @property
    def i_offset(self) -> ModelParameter:
        """
        Settable model parameter: :math:`I_{offset}`
        """
        return self.__i_offset

    @property
    def v_reset(self) -> ModelParameter:
        """
        Settable model parameter: :math:`V_{reset}`
        """
        return self.__v_reset

    @property
    def tau_refrac(self) -> ModelParameter:
        r"""
        Settable model parameter: :math:`\tau_{refrac}`
        """
        return self.__tau_refrac



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union


def _clamp(a: int, b: int, c: int) -> int:
    """
    Force `b` to be between `a` and `c`. `a` must be no larger than `c`.
    """
    return max(a, min(b, c))


def _active_time_for_frequency(frequency: Union[int, float]) -> int:
    if frequency > 0:
        return int(1000000.0 / float(frequency))
    return 0


class MunichIoEthernetProtocol(object):
    """
    Implementation of the Munich robot IO protocol, communicating over
    Ethernet.
    """

    def __init__(self) -> None:
        # Nothing to do here
        pass

    @staticmethod
    def enable_retina() -> bytes:
        """
        :returns: Command to enable the retina.
        """
        return b"E+\n"

    @staticmethod
    def disable_retina() -> bytes:
        """
        :returns: Command to disable the retina.
        """
        return b"E-\n"

    @staticmethod
    def set_retina_transmission(event_format: int) -> bytes:
        """
        :param event_format:
        :returns: Command to set the retina transmission.
        """
        return f"!E{event_format}\n".encode("ascii")

    @staticmethod
    def disable_motor() -> bytes:
        """
        :returns: Command to disable the motor.
        """
        return b"!M-\n"

    @staticmethod
    def enable_motor() -> bytes:
        """
        :returns: Command to enable the motor.
        """
        return b"!M+\n"

    @staticmethod
    def motor_0_permanent_velocity(velocity: int) -> bytes:
        """
        Command to set the motor 0 permanent velocity.

        The value will be restricted to be between -100 and 100.

        :param velocity:
        :returns: Command to set the motor 0 permanent velocity
        """
        return f"!MV0={_clamp(-100, velocity, 100)}\n".encode("ascii")

    @staticmethod
    def motor_1_permanent_velocity(velocity: int) -> bytes:
        """
        Command to set the motor 1 permanent velocity.

        The value will be restricted to be between -100 and 100.

        :param velocity:
        :returns: Command to set the motor 1 permanent velocity.
        """
        return f"!MV1={_clamp(-100, velocity, 100)}\n".encode("ascii")

    @staticmethod
    def motor_0_leaky_velocity(velocity: int) -> bytes:
        """
        Command to set the motor 0 leaky velocity.

        The value will be restricted to be between -100 and 100.

        :param velocity:
        :returns:
        """
        return f"!MVD0={_clamp(-100, velocity, 100)}\n".encode("ascii")

    @staticmethod
    def motor_1_leaky_velocity(velocity: int) -> bytes:
        """
        Command to set the motor 1 leaky velocity.

        The value will be restricted to be between -100 and 100.

        :param velocity:
        :returns: Command to set the motor 1 leaky velocity.
        """
        return f"!MVD1={_clamp(-100, velocity, 100)}\n".encode("ascii")

    @staticmethod
    def led_total_period(total_period: int) -> bytes:
        """
        :param total_period:
        :returns:Command to set the led total time
        """
        return f"!PC={total_period}\n".encode("ascii")

    @staticmethod
    def led_front_active_time(active_time: int) -> bytes:
        """
        :param active_time:
        :returns: Command to set the led front time
        """
        return f"!PC1={active_time}\n".encode("ascii")

    @staticmethod
    def led_back_active_time(active_time: int) -> bytes:
        """
        :param active_time:
        :returns: Command to set the led back time
        """
        return f"!PC0={active_time}\n".encode("ascii")

    @staticmethod
    def led_frequency(frequency: Union[int, float]) -> bytes:
        """
        :param frequency:
        :returns: Command to set the led times based on frequency
        """
        active_time = _active_time_for_frequency(frequency)
        at2 = active_time // 2
        return f"!PC={active_time}\n!PC0={at2}\n!PC1={at2}\n".encode("ascii")

    @staticmethod
    def speaker_frequency(frequency: Union[int, float]) -> bytes:
        """
        :param frequency:
        :returns: Command to set the speaker times based on frequency.
        """
        active_time = _active_time_for_frequency(frequency)
        return f"!PB={active_time}\n!PB0={active_time // 2}\n".encode("ascii")

    @staticmethod
    def speaker_total_period(total_period: int) -> bytes:
        """
        :param total_period:
        :returns: Command to set the speaker total time.
        """
        return f"!PB={total_period}\n".encode("ascii")

    @staticmethod
    def speaker_active_time(active_time: int) -> bytes:
        """
        :param active_time:
        :returns: Command to set the speaker active time.
        """
        return f"!PB0={active_time}\n".encode("ascii")

    @staticmethod
    def laser_frequency(frequency: Union[int, float]) -> bytes:
        """
        :param frequency:
        :returns: Command to set the laser periods based on the frequency.
        """
        active_time = _active_time_for_frequency(frequency)
        return f"!PA={active_time}\n!PA0={active_time // 2}\n".encode("ascii")

    @staticmethod
    def laser_total_period(total_period: int) -> bytes:
        """
        :param total_period:
        :returns: Command to set the laser total period
        """
        return f"!PA={total_period}\n".encode("ascii")

    @staticmethod
    def laser_active_time(active_time: int) -> bytes:
        """
        :param active_time:
        :returns: Command to set the laser active time
        """
        return f"!PA0={active_time}\n".encode("ascii")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_random_stats import AbstractRandomStats
from .random_stats_binomial_impl import RandomStatsBinomialImpl
from .random_stats_exponential_impl import RandomStatsExponentialImpl
from .random_stats_exponential_clipped_impl import (
    RandomStatsExponentialClippedImpl)
from .random_stats_gamma_impl import RandomStatsGammaImpl
from .random_stats_log_normal_impl import RandomStatsLogNormalImpl
from .random_stats_normal_clipped_impl import RandomStatsNormalClippedImpl
from .random_stats_normal_impl import RandomStatsNormalImpl
from .random_stats_poisson_impl import RandomStatsPoissonImpl
from .random_stats_randint_impl import RandomStatsRandIntImpl
from .random_stats_scipy_impl import RandomStatsScipyImpl
from .random_stats_uniform_impl import RandomStatsUniformImpl
from .random_stats_vonmises_impl import RandomStatsVonmisesImpl

__all__ = ["AbstractRandomStats", "RandomStatsBinomialImpl",
           "RandomStatsExponentialImpl", "RandomStatsGammaImpl",
           "RandomStatsLogNormalImpl", "RandomStatsNormalClippedImpl",
           "RandomStatsNormalImpl", "RandomStatsPoissonImpl",
           "RandomStatsRandIntImpl", "RandomStatsScipyImpl",
           "RandomStatsUniformImpl", "RandomStatsVonmisesImpl",
           "RandomStatsExponentialClippedImpl"]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, List, Union, Tuple
from spinn_utilities.overrides import overrides
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spynnaker.pyNN.extra_algorithms.splitter_components import (
    SplitterPopulationVertex)
from spynnaker.pyNN.models.neuron import (
    PopulationVertex, AbstractPyNNNeuronModelStandard)
from spynnaker.pyNN.models.defaults import (
    default_initial_values, default_parameters)
from spynnaker.pyNN.models.neuron.input_types import InputTypeCurrent
from spynnaker.pyNN.models.neuron.implementations import NeuronImplStandard
from spynnaker.pyNN.models.neuron.neuron_models import (
    NeuronModelLeakyIntegrateAndFire)
from spynnaker.pyNN.models.neuron.synapse_types import SynapseTypeExponential
from .abstract_ethernet_translator import AbstractEthernetTranslator
from .abstract_multicast_controllable_device import (
    AbstractMulticastControllableDevice)
from .external_device_lif_control_vertex import ExternalDeviceLifControlVertex
from .threshold_type_multicast_device_control import (
    ThresholdTypeMulticastDeviceControl)


class ExternalDeviceLifControl(AbstractPyNNNeuronModelStandard):
    """
    Abstract control module for the PushBot, based on the LIF neuron, but
    without spikes, and using the voltage as the output to the various devices.
    """
    __slots__ = (
        "_create_edges",
        "_devices",
        "_translator")

    @default_initial_values({"v", "isyn_exc", "isyn_inh"})
    @default_parameters({
        "tau_m", "cm", "v_rest", "v_reset", "tau_syn_E", "tau_syn_I",
        "tau_refrac", "i_offset"})
    def __init__(
            self, devices:  List[AbstractMulticastControllableDevice],
            create_edges: bool,
            translator: Optional[AbstractEthernetTranslator] = None,
            # default params for the neuron model type
            tau_m: float = 20.0, cm: float = 1.0, v_rest: float = 0.0,
            v_reset: float = 0.0, tau_syn_E: float = 5.0,
            tau_syn_I: float = 5.0, tau_refrac: float = 0.1,
            i_offset: float = 0.0, v: float = 0.0, isyn_exc: float = 0.0,
            isyn_inh: float = 0.0):
        """
        :param devices:
            The AbstractMulticastControllableDevice instances to be controlled
            by the population
        :param create_edges:
            True if edges to the devices should be added by this device (set
            to False if using the device over Ethernet using a translator)
        :param translator:
            Translator to be used when used for Ethernet communication.  Must
            be provided if the device is to be controlled over Ethernet.
        :param tau_m: (defaulted LIF neuron parameter)
        :param cm: (defaulted LIF neuron parameter)
        :param v_rest: (defaulted LIF neuron parameter)
        :param v_reset: (defaulted LIF neuron parameter)
        :param tau_syn_E: (defaulted LIF neuron parameter)
        :param tau_syn_I: (defaulted LIF neuron parameter)
        :param tau_refrac: (defaulted LIF neuron parameter)
        :param i_offset: (defaulted LIF neuron parameter)
        :param v: (defaulted LIF neuron state variable initial value)
        :param isyn_exc:
            (defaulted LIF neuron state variable initial value)
        :param isyn_inh:
            (defaulted LIF neuron state variable initial value)
        """
        if not devices:
            raise ConfigurationException("No devices specified")

        neuron_model = NeuronModelLeakyIntegrateAndFire(
            v, v_rest, tau_m, cm, i_offset, v_reset, tau_refrac)
        synapse_type = SynapseTypeExponential(
            tau_syn_E, tau_syn_I, isyn_exc, isyn_inh)
        input_type = InputTypeCurrent()
        threshold_type = ThresholdTypeMulticastDeviceControl(devices)

        self._devices = devices
        self._translator = translator
        self._create_edges = create_edges

        super().__init__(
            model_name="ExternalDeviceLifControl",
            binary="external_device_lif_control.aplx",
            neuron_model=neuron_model, input_type=input_type,
            synapse_type=synapse_type, threshold_type=threshold_type)

    @overrides(AbstractPyNNNeuronModelStandard.create_vertex)
    def create_vertex(
            self, n_neurons: int, label: str, *,
            spikes_per_second: Optional[float] = None,
            ring_buffer_sigma: Optional[float] = None,
            max_expected_summed_weight: Optional[List[float]] = None,
            incoming_spike_buffer_size: Optional[int] = None,
            drop_late_spikes: Optional[bool] = None,
            splitter: Optional[SplitterPopulationVertex] = None,
            seed: Optional[int] = None, n_colour_bits: Optional[int] = None,
            n_steps_per_timestep: int = 1,
            neurons_per_core: Optional[Union[int, Tuple[int, ...]]] = None,
            n_synapse_cores: Optional[int] = None,
            allow_delay_extensions: Optional[bool] = None) -> PopulationVertex:
        if n_neurons != len(self._devices):
            raise ConfigurationException(
                "Number of neurons does not match number of "
                f"devices in {label}")
        model = self._model
        assert isinstance(model, NeuronImplStandard)
        model.n_steps_per_timestep = n_steps_per_timestep
        if neurons_per_core is None:
            neurons_per_core = \
                self.get_model_max_atoms_per_dimension_per_core()
        if n_synapse_cores is None:
            n_synapse_cores = self.get_model_n_synapse_cores()
        if allow_delay_extensions is None:
            allow_delay_extensions = self.get_model_allow_delay_extensions()
        return ExternalDeviceLifControlVertex(
            devices=self._devices, create_edges=self._create_edges,
            max_atoms_per_core=neurons_per_core,
            n_synapse_cores=n_synapse_cores,
            allow_delay_extensions=allow_delay_extensions,
            neuron_impl=model, pynn_model=self, translator=self._translator,
            spikes_per_second=spikes_per_second, label=label,
            ring_buffer_sigma=ring_buffer_sigma,
            max_expected_summed_weight=max_expected_summed_weight,
            incoming_spike_buffer_size=incoming_spike_buffer_size,
            drop_late_spikes=drop_late_spikes, splitter=splitter, seed=seed,
            n_colour_bits=n_colour_bits)
